import numpy as np

from foxes.core import WakeK
from foxes.models.wake_models.top_hat import TopHatWakeModel
import foxes.variables as FV
import foxes.constants as FC


class CrespoHernandezTIWake(TopHatWakeModel):
    """
    The Crespo and Hernandez TI empirical correlation

    Notes
    -----
    Reference:
    "Turbulence characteristics in wind-turbine wakes"
    A. Crespo, J. Hernandez
    https://doi.org/10.1016/0167-6105(95)00033-X

    For the wake diameter we use Eqns. (17), (15), (4), (5) from
            doi:10.1088/1742-6596/625/1/012039

    Attributes
    ----------
    a_near: float
        Model parameter
    a_far: float
        Model parameter
    e1: float
        Model parameter
    e2: float
        Model parameter
    e3: float
        Model parameter
    use_ambti: bool
        Flag for using ambient TI instead of local
        wake corrected TI
    sbeta_factor: float
        Factor multiplying sbeta
    near_wake_D: float
        The near wake distance in units of D,
        calculated from TI and ct if None
    wake_k: foxes.core.WakeK
        Handler for the wake growth parameter k

    :group: models.wake_models.ti

    """

    def __init__(
        self,
        superposition,
        use_ambti=False,
        sbeta_factor=0.25,
        near_wake_D=None,
        a_near=0.362,
        a_far=0.73,
        e1=0.83,
        e2=-0.0325,
        e3=-0.32,
        induction="Betz",
        **wake_k,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The TI wake superposition.
        k: float, optional
            The wake growth parameter k. If not given here
            it will be searched in the farm data.
        use_ambti: bool
            Flag for using ambient TI instead of local
            wake corrected TI
        sbeta_factor: float
            Factor multiplying sbeta
        near_wake_D: float, optional
            The near wake distance in units of D,
            calculated from TI and ct if not given here
        a_near: float
            Model parameter
        a_far: float
            Model parameter
        e1: float
            Model parameter
        e2: float
            Model parameter
        e3: float
            Model parameter
        k_var: str
            The variable name for k
        induction: foxes.core.AxialInductionModel or str
            The induction model
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.TI: superposition}, induction=induction)

        self.a_near = a_near
        self.a_far = a_far
        self.e1 = e1
        self.e2 = e2
        self.e3 = e3
        self.use_ambti = use_ambti
        self.sbeta_factor = sbeta_factor
        self.near_wake_D = near_wake_D
        self.wake_k = WakeK(**wake_k)

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.TI]}, induction={iname}, "
        s += self.wake_k.repr() + ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.wake_k]

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_rpoints, ...)

        """
        return {FV.TI: np.zeros_like(tdata[FC.TARGETS][..., 0])}

    def calc_wake_radius(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        ct,
    ):
        """
        Calculate the wake radius, depending on x only (not r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_states, n_targets)

        Returns
        -------
        wake_r: numpy.ndarray
            The wake radii, shape: (n_states, n_targets)

        """
        # get D:
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        # get k:
        k = self.wake_k(
            FC.STATE_TARGET,
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        # calculate:
        a = self.induction.ct2a(ct)
        beta = (1 - a) / (1 - 2 * a)
        radius = 2 * (k * x + self.sbeta_factor * np.sqrt(beta) * D)

        return radius

    def calc_centreline(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        x,
        wake_r,
        ct,
    ):
        """
        Calculate centre line results of wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)
        x: numpy.ndarray
            The x values, shape: (n_st_sel,)
        wake_r: numpy.ndarray
            The wake radii, shape: (n_st_sel,)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_st_sel,)

        Returns
        -------
        cl_del: dict
            The centre line wake deltas. Key: variable name str,
            varlue: numpy.ndarray, shape: (n_st_sel,)

        """
        # prepare:
        n_targts = np.sum(st_sel)
        TI = FV.AMB_TI if self.use_ambti else FV.TI

        # get D:
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )[st_sel]

        # get TI:
        ti = self.get_data(
            TI,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )[st_sel]

        # calculate induction factor:
        twoa = 2 * self.induction.ct2a(ct)

        # prepare output:
        wake_deltas = np.zeros(n_targts, dtype=FC.DTYPE)

        # calc near wake length, if not given
        if self.near_wake_D is None:
            near_wake_D = (
                2**self.e1
                * self.a_near
                / (self.a_far * ti**self.e2)
                * twoa ** (1 - self.e1)
            ) ** (1 / self.e3)
        else:
            near_wake_D = self.near_wake_D

        # calc near wake:
        sel = x < near_wake_D * D
        if np.any(sel):
            wake_deltas[sel] = self.a_near * twoa[sel]

        # calc far wake:
        if np.any(~sel):
            # calculate delta:
            #
            # Note the sign flip of the exponent ti[~sel]**(-0.0325)
            # compared to the original paper. This was found in
            # https://doi.org/10.1016/j.jweia.2018.04.010, Eq. (46)
            # Without this flip the near and far wake areas are not
            # smoothly connected.
            #
            wake_deltas[~sel] = (
                self.a_far
                * (twoa[~sel] / 2) ** self.e1
                * ti[~sel] ** self.e2
                * (x[~sel] / D[~sel]) ** self.e3
            )

        return {FV.TI: wake_deltas}



import numpy as np
import pandas as pd
from xarray import Dataset
from numbers import Number

from foxes.core import States
import foxes.constants as FC
import foxes.variables as FV

def _get_profiles(coords, fields, dims, ovars, fixval, verbosity):
    """Read ABL profiles information
    :group: input.windio
    """
    profiles = {}
    if FV.Z0 in fields:
        if FV.H not in fields and verbosity > 0:
            print(f"Ignoring '{FV.Z0}', since no reference_height found. No ABL profile activated.")
        elif FV.MOL in fields:
            ovars.append(FV.MOL)
            fixval[FV.H] = fields[FV.H]
            profiles = {FV.WS: "ABLLogWsProfile"}
        else:
            fixval[FV.H] = fields[FV.H]
            profiles = {FV.WS: "ABLLogNeutralWsProfile"}
    elif FV.H in fields and verbosity > 0:
        print(f"Ignoring '{FV.H}', since no '{FV.Z0}' data found. No ABL profile activated.")
    if len(profiles) and verbosity > 2:
        print(f"        Selecting ABL profile '{profiles[FV.WS]}', {FV.H} = {fields[FV.H]} m")
            
    return profiles
    
def _get_SingleStateStates(coords, fields, dims, states_dict, 
                           ovars, fixval, profiles, verbosity):
    """Try to generate single state parameters
    :group: input.windio
    """
    for c in coords:
        if not isinstance(c, Number):
            return False

    if verbosity > 2:
        print("        selecting class 'SingleStateStates'")

    smap = {FV.WS: "ws", FV.WD: "wd", FV.TI: "ti", FV.RHO: "rho"}

    data = {smap[v]: d for v, d in fixval.items()}
    for v, d in coords.items():
        if v in smap:
            data[smap[v]] = d
        elif verbosity > 1:
            print(f"        ignoring coord '{v}'")
    for v, d in fields.items():
        if v in smap and len(dims[v]) == 0:
            data[smap[v]] = d
        elif verbosity > 1:
            print(f"        ignoring field '{v}' with dims {dims[v]}")

    sdata = pd.DataFrame(index=coords[FC.TIME], data=data)
    sdata.index.name = FC.TIME
    states_dict.update(
        dict(
            states_type="SingleStateStates",
            profiles=profiles,
            **data,
        )
    )
    return True

def _get_Timeseries(coords, fields, dims, states_dict, 
                    ovars, fixval, profiles, verbosity):
    """Try to generate time series parameters
    :group: input.windio
    """
    if len(coords) == 1 and FC.TIME in coords:
        if verbosity > 2:
            print("        selecting class 'Timeseries'")

        data = {}
        fix = {}
        for v, d in fields.items():
            if dims[v] == (FC.TIME,):
                data[v] = d
            elif len(dims[v]) == 0:
                fix[v] = d
            elif verbosity > 2:
                print(f"        ignoring field '{v}' with dims {dims[v]}")
        fix.update({v: d for v, d in fixval.items() if v not in data})

        sdata = pd.DataFrame(index=coords[FC.TIME], data=data)
        sdata.index.name = FC.TIME
        states_dict.update(
            dict(
                states_type="Timeseries",
                data_source=sdata,
                output_vars=ovars,
                fixed_vars=fix,
                profiles=profiles,
            )
        )
        return True
    return False

def _get_MultiHeightNCTimeseries(coords, fields, dims, states_dict, 
                    ovars, fixval, profiles, verbosity):
    """Try to generate time series parameters
    :group: input.windio
    """
    if len(coords) == 2 and FC.TIME in coords and FV.H in coords:
        if verbosity > 2:
            print("        selecting class 'MultiHeightNCTimeseries'")
            
        if len(profiles) and verbosity > 0:
            print(f"Ignoring profile '{profiles[FV.WS]}' for states class 'MultiHeightNCTimeseries'")

        data = {}
        fix = {}
        for v, d in fields.items():
            if dims[v] == (FC.TIME, FV.H):
                data[v] = ((FC.TIME, FV.H), d)
            if dims[v] == (FV.H, FC.TIME):
                data[v] = ((FC.TIME, FV.H), np.swapaxes(d, 0, 1))
            elif len(dims[v]) == 0:
                fix[v] = d
            elif verbosity > 2:
                print(f"        ignoring field '{v}' with dims {dims[v]}")
        fix.update({v: d for v, d in fixval.items() if v not in data})

        sdata = Dataset(coords=coords, data_vars=data)
        states_dict.update(
            dict(
                states_type="MultiHeightNCTimeseries",
                h_coord=FV.H,
                format_times_func=None,
                data_source=sdata,
                output_vars=ovars,
                fixed_vars=fix,
            )
        )
        return True
    return False

def get_states(coords, fields, dims, verbosity=1):
    """
    Reads states parameters from windio input

    Parameters
    ----------
    coords: dict
        The coordinates data
    fields: dict
        The fields data
    dims: dict
        The dimensions data
    verbosity: int
        The verbosity level

    Returns
    -------
    states: foxes.core.States
        The states object

    :group: input.windio

    """
    if verbosity > 2:
        print("      Preparing states")
        
    ovars = [FV.WS, FV.WD, FV.TI, FV.RHO]
    fixval = {FV.TI: 0.05, FV.RHO: 1.225}  
    profiles = _get_profiles(coords, fields, dims, ovars, fixval, verbosity)

    states_dict = {}
    if _get_SingleStateStates(
        coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
    ) or _get_Timeseries(
        coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
    ) or _get_MultiHeightNCTimeseries(
        coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
    ):
        return States.new(**states_dict)
    else:
        raise ValueError(
            f"Failed to create states for coords {list(coords.keys())} and fields {list(fields.keys())} with dims {dims}"
        )



"""
General utilities.
"""

from .wind_dir import wd2uv, wd2wdvec, wdvec2wd, uv2wd, delta_wd
from .pandas_utils import PandasFileHelper
from .xarray_utils import write_nc
from .subclasses import all_subclasses
from .dict import Dict
from .factory import Factory, FDict, WakeKFactory
from .data_book import DataBook
from .cubic_roots import cubic_roots
from .geopandas_utils import read_shp, shp2csv, read_shp_polygons, shp2geom2d
from .load import import_module, load_module
from .exec_python import exec_python
from .regularize import sqrt_reg
from .windrose_plot import TabWindroseAxes
from .tab_files import read_tab_file
from .random_xy import random_xy_square

from . import two_circles
from . import abl
from . import runners
from . import geom2d



import foxes.variables as FV
from foxes.core import PointDataModel


class SetAmbPointResults(PointDataModel):
    """
    This model copies point results to ambient results.

    Attributes
    ----------
    pvars: list of str
        The point variables to be treated
    vars: list of str
        The variables to be copied to output

    :group: algorithms.downwind.models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__()
        self.pvars = None
        self.vars = None

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        self.pvars = algo.states.output_point_vars(algo)
        self.vars = [v for v in self.pvars if v in FV.var2amb]
        super().initialize(algo, verbosity)

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.var2amb[v] for v in self.vars]

    def calculate(self, algo, mdata, fdata, pdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        pdata: foxes.core.Data
            The point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """
        for v in self.vars:
            pdata[FV.var2amb[v]] = pdata[v].copy()
        return {v: pdata[v] for v in self.output_point_vars(algo)}



import numpy as np

from foxes.utils import random_xy_square
from foxes.core import Turbine


def add_random(
    farm,
    n_turbines,
    min_dist,
    centre=[0, 0],
    seed=None,
    verbosity=1,
    **turbine_parameters,
):
    """
    Add turbines that lie randomly within a square

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    n_turbines: int
        The number of turbines
    min_dist: float
        The minimal distance between turbines
    centre: array-like
        The (x, y) coordinates of the mean
    seed: int, optional
        The random seed
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Additional parameters are forwarded to the WindFarm.add_turbine().

    :group: input.farm_layout

    """
    xy = random_xy_square(n_turbines, min_dist, seed=seed, verbosity=verbosity)
    xy += np.array(centre)[None, :]

    for i in range(len(xy)):
        farm.add_turbine(
            Turbine(
                xy=xy[i],
                **turbine_parameters,
            ),
            verbosity=verbosity,
        )



"""
Partial wake models.
"""

from .centre import PartialCentre
from .rotor_points import RotorPoints
from .top_hat import PartialTopHat
from .axiwake import PartialAxiwake
from .segregated import PartialSegregated
from .grid import PartialGrid



import numpy as np
from .neutral import logz as lgz


def logz(height, z0):
    """
    Calculates the log factor for
    wind speed profiles.

    Parameters
    ----------
    height: float
        The evaluation height
    z0: float
        The roughness length

    Returns
    -------
    lz: float
        The log factor

    :group: utils.abl.stable

    """
    return lgz(height, z0)


def psi(height, mol):
    """
    The Psi function

    Parameters
    ----------
    height: float
        The height value
    mol: float
        The Monin-Obukhov height

    Returns
    -------
    psi: float
        The Psi function value

    :group: utils.abl.stable

    """
    h = np.minimum(height, np.abs(mol))
    return -5.0 * h / mol


def ustar(ws_ref, h_ref, z0, mol, kappa=0.41):
    """
    Calculates the friction velocity,
    based on reference data.

    Parameters
    ----------
    ws_ref: float
        The reference wind speed
    h_ref: float
        The reference height
    z0: float
        The roughness length
    mol: float
        The Monin-Obukhov height
    kappa: float
        The van-Karman constant

    Returns
    -------
    ustar: float
        The friction velocity

    :group: utils.abl.stable

    """
    return ws_ref * kappa / (logz(h_ref, z0) - psi(h_ref, mol))


def calc_ws(height, z0, ustar, psi, kappa=0.41):
    """
    Calculate wind speeds at given height

    Parameters
    ----------
    height: float
        The evaluation height
    z0: float
        The roughness length
    ustar: float
        The friction velocity
    psi: float
        The Psi function values
    kappa: float
        The van-Karman constant

    Returns
    -------
    ws: float
        The wind speed

    :group: utils.abl.stable

    """
    return ustar / kappa * (logz(height, z0) - psi)



import numpy as np
import pandas as pd
import xarray as xr

from foxes.core import TurbineModel
from foxes.utils import PandasFileHelper
import foxes.constants as FC


class LookupTable(TurbineModel):
    """
    Calculates the data by interpolation of
    lookup-table data

    Attributes
    ----------
    data_source: str or pandas.DataFrame
        The lookup-table data
    input_vars: list of str
        The foxes input variables
    output_vars: list of str
        The foxes output variables
    varmap: dict
        Mapping from foxes variable names
        to column names in the data_source

    :group: models.turbine_models

    """

    def __init__(
        self,
        data_source,
        input_vars,
        output_vars,
        varmap={},
        pd_file_read_pars={},
        xr_interp_args={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            The lookup-table data
        input_vars: list of str
            The foxes input variables
        output_vars: list of str
            The foxes output variables
        varmap: dict
            Mapping from foxes variable names
            to column names in the data_source
        pd_file_read_pars: dict
            Parameters for pandas file reading
        xr_interp_args: dict
            Parameters for xarray interpolation method
        kwargs: dict, optional
            Additional parameters, added as default
            values if not in data

        """
        super().__init__()

        self.data_source = data_source
        self.input_vars = input_vars
        self.output_vars = output_vars
        self.varmap = varmap

        self._rpars = pd_file_read_pars
        self._xargs = xr_interp_args
        self._data = None

        for v, d in kwargs.items():
            if v not in input_vars:
                raise KeyError(
                    f"{self.name}: Default input parameter '{v}' not in list of inputs {input_vars}"
                )
            setattr(self, v, d)

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.output_vars

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        if self._data is None:
            if isinstance(self.data_source, pd.DataFrame):
                data = self.data_source
            else:
                if verbosity > 0:
                    print(f"{self.name}: Reading file {self.data_source}")
                data = PandasFileHelper.read_file(self.data_source, **self._rpars)

            if verbosity > 0:
                print(f"{self.name}: Preparing interpolation data")
            if len(self.varmap):
                data = data.rename(columns={c: v for v, c in self.varmap.items()})
            data = data[self.input_vars + self.output_vars]
            data.sort_values(by=self.input_vars, inplace=True)
            coords = {
                v: np.asarray(data[v].unique(), dtype=FC.DTYPE) for v in self.input_vars
            }

            dvars = {}
            for oname in self.output_vars:
                pivot_matrix = data.pivot_table(index=self.input_vars, values=[oname])
                dvars[oname] = (
                    self.input_vars,
                    pivot_matrix.to_numpy(FC.DTYPE).reshape(
                        pivot_matrix.index.levshape
                    ),
                )

            self._data = xr.Dataset(coords=coords, data_vars=dvars)

            if verbosity > 1:
                print()
                print(self._data)
                print()

        return super().load_data(algo, verbosity)

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        data = {
            v: self.get_data(
                self.input_vars[0],
                FC.STATE_TURBINE,
                lookup="fs",
                fdata=fdata,
                upcast=True,
            )[st_sel]
            for v in self.input_vars
        }
        dims = {
            v: ("_z") if len(data[v].shape) == 1 else ("_z", "_u")
            for v in self.input_vars
        }
        indata = {
            v: xr.DataArray(
                data[v],
                dims=dims[v],
            )
            for v in self.input_vars
        }
        del data, dims

        odata = self._data.interp(**indata, **self._xargs)

        out = {}
        for v in self.output_vars:
            out[v] = fdata[v]
            out[v][st_sel] = odata[v]

        return out



import numpy as np

from foxes.core import TurbineModel


class Calculator(TurbineModel):
    """
    Calculates variables based on given functions.

    Attributes
    ----------
    in_vars: list of str
        The input farm vairables
    out_vars: list of str
        The output variables
    func: Function
        The function: f(in0, in1, ..., stsel) -> (out0, ou1, ...)
        where inX and outY are numpy.ndarrays and
        st_sel is the state-turbine selection slice or array.
        All arrays have shape (n_states, n_turbines).

    :group: models.turbine_models

    """

    def __init__(self, in_vars, out_vars, func, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        in_vars: list of str
            The input farm vairables
        out_vars: list of str
            The output variables
        func: Function
            The function: f(in0, in1, ..., stsel) -> (out0, ou1, ...)
            where inX and outY are numpy.ndarrays and
            st_sel is the state-turbine selection slice or array.
            All arrays have shape (n_states, n_turbines).
        kwargs: dict, optional
            Additional arguments for TurbineModel

        """
        super().__init__(**kwargs)
        self.in_vars = in_vars
        self.out_vars = out_vars
        self.func = func

    def __repr__(self):
        a = f"{self.in_vars}, {self.out_vars}"
        return f"{type(self).__name__}({a})"

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.out_vars

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        ins = [fdata[v] if v in fdata else mdata[v] for v in self.in_vars]
        outs = self.func(*ins, st_sel=st_sel)

        return {v: outs[vi] for vi, v in enumerate(self.out_vars)}



import numpy as np
from scipy.spatial.distance import cdist


def random_xy_square(
    n,
    min_dist=500,
    xmax_ini=None,
    growth=1.02,
    seed=None,
    verbosity=1,
):
    """
    Creates random xy positions within a square,
    with mean (0, 0)

    Parameters
    ----------
    n: int
        The number of positions
    min_dist: float
        The minimal distance between any two positions
    xmax_ini: float, optional
        The initial maximal distance of any coordinates
    growth: float
        The growth factor of the initial radius, must be
        greater 1
    seed: int, optional
        The random seed
    verbosity: int
        The verbosity level. 0 = silent

    Returns
    -------
    xy: numpy.ndarray
        The positions, shape: (n, 2)

    :group: utils

    """
    if seed:
        np.random.seed(seed)

    xmax = np.sqrt(n) * min_dist if xmax_ini is None else xmax_ini
    xy = np.random.uniform(0, xmax, (n, 2))
    while True:
        dists = cdist(xy, xy)
        np.fill_diagonal(dists, np.inf)
        sel = np.unique(np.where(dists < min_dist)[0])
        if not len(sel):
            break
        if verbosity > 0:
            print(f"Re-generating coordinates: {len(sel)}, xmax = {xmax:.1f}")
        xmax *= growth
        xy[sel] = np.random.uniform(0, xmax, (len(sel), 2))
    return xy - np.mean(xy, axis=0)[None, :]



from foxes.core.states import States


class SeqState(States):
    """
    A single state during sequential iteration, just serving
    as a structural placeholder

    Parameters
    ----------
    states: foxes.core.States
        The original states set

    :group: algorithms.sequential.models

    """

    def __init__(self, states):
        """
        Constructor.

        Attributes
        ----------
        states: foxes.core.States
            The original states set

        """
        super().__init__()
        self.states = states

        # updated by SequentialIter:
        self._size = states.size()
        self._weight = None
        self._indx = None
        self._counter = None

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.states]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        super().initialize(algo, verbosity, force)
        self._size = self.states.size()

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._size

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return [self._indx] if self._size == 1 else self.states.index()

    @property
    def counter(self):
        """
        The current index counter

        Returns
        -------
        i: int
            The current index counter

        """
        return self._counter

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return self._weight[None, :] if self._size == 1 else self.states.weights(algo)

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.states.output_point_vars(algo)

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        return self.states.calculate(algo, mdata, fdata, tdata)



from foxes.core import VerticalProfile
from foxes.utils.abl import neutral
import foxes.constants as FC
import foxes.variables as FV


class ABLLogNeutralWsProfile(VerticalProfile):
    """
    The neutral ABL wind speed log profile.

    Attributes
    ----------
    ustar_input: bool
        Flag for using ustar as an input

    :group: models.vertical_profiles

    """

    def __init__(self, *args, ustar_input=False, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Additional arguments for VerticalProfile
        ustar_input: bool
            Flag for using ustar as an input
        kwargs: dict, optional
            Additional arguments for VerticalProfile

        """
        super().__init__(*args, **kwargs)
        self.ustar_input = ustar_input

    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        if self.ustar_input:
            return [FV.USTAR, FV.Z0]
        else:
            return [FV.WS, FV.H, FV.Z0]

    def calculate(self, data, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        data: dict
            The input data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """
        z0 = data[FV.Z0]
        if self.ustar_input:
            ustar = data[FV.USTAR]
        else:
            h0 = data[FV.H]
            ws = data[FV.WS]
            ustar = neutral.ustar(ws, h0, z0, kappa=FC.KAPPA)

        return neutral.calc_ws(heights, z0, ustar, kappa=FC.KAPPA)



"""
Turbine models.
"""

from .kTI_model import kTI
from .set_farm_vars import SetFarmVars
from .thrust2ct import Thrust2Ct
from .power_mask import PowerMask
from .sector_management import SectorManagement
from .table_factors import TableFactors
from .yaw2yawm import YAW2YAWM
from .yawm2yaw import YAWM2YAW
from .calculator import Calculator
from .rotor_centre_calc import RotorCentreCalc
from .lookup_table import LookupTable



from .dict import Dict
import foxes.variables as FV


class Factory:
    """
    Constructs objects from a choice of allowed
    constructor parameters

    Attributes
    ----------
    base: class
        The class of which objects are to be created
    name_template: str
        The name template, e.g. 'name_<A>_<B>_<C>' for
        variables A, B, C
    args: tuple
        Fixed arguments for the base class
    kwargs: dict
        Fixed arguments for the base class
    var2arg: dict
        Mapping from variable to constructor argument
    hints: dict
        Hints for print_toc, only for variables for which the
        options are functions or missing
    options: dict
        For each variable, e.g. A, B or C, the list or dict
        or function that maps a str to the actual value

    :group: utils

    """

    def __init__(
        self,
        base,
        name_template,
        args=(),
        kwargs={},
        var2arg={},
        hints={},
        **options,
    ):
        """
        Constructor.

        Parameters
        ----------
        base: class
            The class of which objects are to be created
        name_template: str
            The name template, e.g. 'name_<A>_<B>_<C>' for
            variables A, B, C
        args: tuple
            Fixed arguments for the base class
        kwargs: dict
            Fixed arguments for the base class
        var2arg: dict
            Mapping from variable to constructor argument
        hints: dict
            Hints for print_toc, only for variables for which the
            options are functions or missing
        options: dict
            For each variable, e.g. A, B or C, the list or dict
            or function that maps a str to the actual value

        """
        self.base = base
        self.name_template = name_template
        self.args = args
        self.kwargs = kwargs
        self.var2arg = var2arg
        self.hints = hints

        self._vars = []
        self._pre = []
        parts = name_template.split(">")

        for i, p in enumerate(parts):
            if i < len(parts) - 1:
                parts2 = p.split("<")
                if len(parts2) != 2:
                    raise ValueError(
                        f"Factory '{name_template}': incomplete pattern brackets '<..>' between variables, e.g. '_'"
                    )
                if i > 0 and len(parts2[0]) == 0:
                    raise ValueError(
                        f"Factory '{name_template}': Missing seperator like '_' in template between variables '{self._vars[-1]}' and '{parts[1]}'"
                    )
                self._pre.append(parts2[0])
                self._vars.append(parts2[1])
            else:
                self._pre.append(p)

        if len(self.variables) > 1:
            for vi, v in enumerate(self.variables):
                p = self._pre[vi]
                if vi < len(self.variables) - 1 and p == "":
                    raise ValueError(
                        f"Factory '{name_template}': Require indicator before variable '{v}' in template, e.g. '{v}<{v}>'"
                    )

        self.options = Dict(name=f"{self._pre[0]}_options")
        for v, o in options.items():
            if v not in self.variables:
                raise KeyError(
                    f"Factory '{name_template}': Variable '{v}' found in options, but not in template"
                )
            if isinstance(o, list) or isinstance(o, tuple):
                o = {str(k): k for k in o}
            if isinstance(o, dict):
                for k in o.keys():
                    if not isinstance(k, str):
                        raise TypeError(
                            f"Factory '{name_template}': Found option for variable '{v}' that is not a str, {k}"
                        )
                self.options[v] = Dict(name=f"{self._pre[0]}_options_{v}", **o)
            elif hasattr(o, "__call__"):
                self.options[v] = o
            else:
                raise ValueError(
                    f"Factory '{name_template}': Variable '{v}' has option of type '{type(v).__name__}'. Only list, tuple, dict or function are supported"
                )

    @property
    def name_prefix(self):
        """
        The beginning of the name template

        Returns
        -------
        nbase: str
            The beginning of the name template

        """
        return self._pre[0]

    @property
    def name_suffix(self):
        """
        The ending of the name template

        Returns
        -------
        nbase: str
            The ending of the name template

        """
        return self._pre[-1]

    @property
    def variables(self):
        """
        The list of variables

        Returns
        -------
        vrs: list of str
            The variables

        """
        return self._vars

    def __str__(self):
        """String representation"""
        s = f"{self.name_template}: {self.base.__name__} with"
        for k, d in self.kwargs.items():
            s += f"\n  {k}={d}"
        for v in self.variables:
            if v in self.options and isinstance(self.options[v], dict):
                s += f"\n  {v} from {list(self.options[v])}"
            else:
                s += f"\n  {v}={self.hints.get(v, '(value)')}"
        return s

    def check_match(self, name):
        """
        Tests if a name matches the template

        Parameters
        ----------
        name: str
            The name to be checked

        Returns
        -------
        success: bool
            True if the template is matched

        """
        data_str = name
        for vi in range(len(self.variables)):
            p = self._pre[vi]
            i = data_str.find(p)
            j = i + len(p)
            if i < 0 or len(data_str) <= j:
                return False
            data_str = data_str[j:]

            q = self._pre[vi + 1]
            if q != "":
                i = data_str.find(q)
                j = i + len(q)
                if i < 0 or len(data_str) <= j:
                    return False
            else:
                data_str = ""

        return True

    def construct(self, name):
        """
        Create an object of the base class.

        Parameters
        ----------
        name: str
            The name, matching the template

        Returns
        -------
        obj: object
            The instance of the base class

        """
        j = 0
        wlist = []
        for pi, p in enumerate(self._pre):
            if len(p) > 0:
                i = name[j:].find(p)
                if i < 0 or (pi == 0 and i > 0):
                    raise ValueError(
                        f"Factory '{self.name_template}': Name '{name}' not matching template"
                    )
                w = name[j : j + i]
                j += i + len(p)
            else:
                w = name[j:]
            if pi > 0:
                wlist.append(w)

        kwargs = {}
        for vi, v in enumerate(self.variables):
            w = self.var2arg.get(v, v)
            data = wlist[vi]
            if v in self.options:
                o = self.options[v]
                if hasattr(o, "__call__"):
                    kwargs[w] = o(data)
                else:
                    kwargs[w] = self.options[v][data]
            else:
                kwargs[w] = data

        kwargs.update(self.kwargs)

        return self.base(*self.args, **kwargs)


class WakeKFactory:
    """
    A factory that automatically handles
    wake_k parameters

    Attributes
    ----------
    factories: list of Factory
        The individual factories

    :group: utils

    """

    def __init__(self, base, name_template, *args, hints={}, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        base: class
            The class of which objects are to be created
        name_template: str
            The name template, e.g. 'name_<A>_<B>_<C>' for
            variables A, B, C. Indicate wake_k part by '_[wake_k]'
        args: tuple, optional
            Additional arguments for Factory
        hints: dict
            Hints for print_toc, only for variables for which the
            options are functions or missing
        kwargs: dict
            Additional arguments for Factory

        """
        self._base = base
        self._kwargs = kwargs
        self._template0 = name_template
        self.factories = []

        i0 = name_template.find("_[wake_k]")
        i1 = i0 + len("_[wake_k]")
        kw = kwargs.pop("kwargs", {})

        if i0 < 0:
            raise ValueError(
                f"String '_[wake_k]' not found in name template '{name_template}'"
            )

        # add case ka, kb:
        t = name_template[:i0] + "_ka<ka>_kb<kb>"
        if len(name_template) > i1:
            t += name_template[i1:]
        h = hints.copy()
        h["ka"] = "(Value, e.g. 04 for 0.4)"
        h["kb"] = "(Value, e.g. 001 for 0.01)"
        kw["ti_var"] = FV.TI
        self.factories.append(
            Factory(
                base,
                t,
                *args,
                hints=h,
                kwargs=kw.copy(),
                **kwargs,
                ka=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
                kb=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            )
        )

        # add case ambient ka, kb:
        t = name_template[:i0] + "_ambka<ka>_kb<kb>"
        if len(name_template) > i1:
            t += name_template[i1:]
        h = hints.copy()
        h["ka"] = "(Value, e.g. 04 for 0.4)"
        h["kb"] = "(Value, e.g. 001 for 0.01)"
        kw["ti_var"] = FV.AMB_TI
        self.factories.append(
            Factory(
                base,
                t,
                *args,
                hints=h,
                kwargs=kw.copy(),
                **kwargs,
                ka=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
                kb=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            )
        )

        # add case ka:
        t = name_template[:i0] + "_ka<ka>"
        if len(name_template) > i1:
            t += name_template[i1:]
        h = hints.copy()
        h["ka"] = "(Value, e.g. 04 for 0.4)"
        kw["ti_var"] = FV.TI
        self.factories.append(
            Factory(
                base,
                t,
                *args,
                hints=h,
                kwargs=kw.copy(),
                **kwargs,
                ka=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            )
        )

        # add case ka:
        t = name_template[:i0] + "_ambka<ka>"
        if len(name_template) > i1:
            t += name_template[i1:]
        h = hints.copy()
        h["ka"] = "(Value, e.g. 04 for 0.4)"
        kw["ti_var"] = FV.AMB_TI
        self.factories.append(
            Factory(
                base,
                t,
                *args,
                hints=h,
                kwargs=kw.copy(),
                **kwargs,
                ka=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            )
        )

        # add case k:
        t = name_template[:i0] + "_k<k>"
        if len(name_template) > i1:
            t += name_template[i1:]
        h = hints.copy()
        h["k"] = "(Value, e.g. 004 for 0.04)"
        kw["ti_var"] = FV.TI
        self.factories.append(
            Factory(
                base,
                t,
                *args,
                hints=h,
                kwargs=kw.copy(),
                **kwargs,
                k=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            )
        )

        # add case without k:
        t = name_template[:i0]
        if len(name_template) > i1:
            t += name_template[i1:]
        kw["ti_var"] = FV.TI
        self.factories.append(
            Factory(
                base,
                t,
                *args,
                hints=hints,
                kwargs=kw.copy(),
                **kwargs,
            )
        )

    def __str__(self):
        """String representation"""
        s = f"{self._template0}: {self._base.__name__} with"
        for k, d in self._kwargs.items():
            s += f"\n  {k}={d}"
        f0 = self.factories[-1]
        for v in f0.variables:
            if v in f0.options and isinstance(f0.options[v], dict):
                s += f"\n  {v} from {list(f0.options[v])}"
            else:
                s += f"\n  {v}={f0.hints.get(v, '(value)')}"
        s += (
            f"\n  [wake_k]=(None or k<k> or ka<ka> or ka<ka>_kb<kb>, e.g. 004 for 0.04)"
        )
        return s


class FDict(Dict):
    """
    A dictionary with factory support

    Attributes
    ----------
    store_created: bool
        Flag for storing created objects
    factories: list of foxes.utils.Factory
        The factories

    :group: utils

    """

    def __init__(self, *args, store_created=True, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Parameters for the base class
        store_created: bool
            Flag for storing created objects
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(*args, **kwargs)
        self.store_created = store_created
        self.factories = []

    def add_factory(self, *args, factory=None, **kwargs):
        """
        Adds a Factory object.

        Parameters
        ----------
        args: tuple, optional
            Parameters for the Factory constructor
        factory: Factory, optional
            The factory object
        kwargs: dict, optional
            Parameters for the Factory constructor

        """
        f = Factory(*args, **kwargs) if factory is None else factory
        i = len(self.factories)
        for gi in range(len(self.factories) - 1, -1, -1):
            g = self.factories[gi]
            if (
                g.name_prefix == f.name_prefix
                and g.name_suffix == f.name_suffix
                and len(f.variables) > len(g.variables)
            ):
                i = gi

        if i == len(self.factories):
            self.factories.append(f)
        else:
            self.factories.insert(i, f)

    def add_k_factory(self, *args, **kwargs):
        """
        Adds a WakeKFactory.

        Parameters
        ----------
        args: tuple, optional
            Parameters for the Factory constructor
        kwargs: dict, optional
            Parameters for the Factory constructor

        """
        for f in WakeKFactory(*args, **kwargs).factories:
            self.add_factory(factory=f)

    def __contains__(self, key):
        found = super().__contains__(key)
        if not found:
            for f in self.factories:
                if f.check_match(key):
                    return True
        return found

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            for f in self.factories:
                try:
                    obj = f.construct(key)
                    if self.store_created:
                        self[key] = obj
                    return obj
                except ValueError:
                    pass

        k = ", ".join(sorted(list(self.keys())))
        e = f"{self.name}: Cannot find key '{key}', also no factory matches. Known keys: {k}. Known factories: {[f.name_template for f in self.factories]}"
        raise KeyError(e)



import numpy as np

from foxes.core import WakeK
from foxes.models.wake_models.gaussian import GaussianWakeModel
import foxes.variables as FV
import foxes.constants as FC


class TurbOParkWake(GaussianWakeModel):
    """
    The TurbOPark wake model

    Notes
    -----
    Reference:
    "Turbulence Optimized Park model with Gaussian wake profile"
    J G Pedersen, E Svensson, L Poulsen and N G Nygaard
    https://iopscience.iop.org/article/10.1088/1742-6596/2265/2/022063/pdf

    Attributes
    ----------
    sbeta_factor: float
        Factor multiplying sbeta
    c1: float
        Factor from Frandsen turbulence model
    c2: float
        Factor from Frandsen turbulence model
    induction: foxes.core.AxialInductionModel or str
        The induction model
    wake_k: foxes.core.WakeK
        Handler for the wake growth parameter k

    :group: models.wake_models.wind

    """

    def __init__(
        self,
        superposition,
        sbeta_factor=0.25,
        c1=1.5,
        c2=0.8,
        induction="Madsen",
        **wake_k,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind deficit superposition
        sbeta_factor: float
            Factor multiplying sbeta
        c1: float
            Factor from Frandsen turbulence model
        c2: float
            Factor from Frandsen turbulence model
        induction: foxes.core.AxialInductionModel or str
            The induction model
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.WS: superposition})

        self.sbeta_factor = sbeta_factor
        self.c1 = c1
        self.c2 = c2
        self.induction = induction
        self.wake_k = WakeK(**wake_k)

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.WS]}, induction={iname}, "
        s += self.wake_k.repr() + ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.wake_k, self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def calc_amplitude_sigma(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Calculate the amplitude and the sigma,
        both depend only on x (not on r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)

        Returns
        -------
        amsi: tuple
            The amplitude and sigma, both numpy.ndarray
            with shape (n_st_sel,)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        # get ct:
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # select targets:
        st_sel = (x > 1e-5) & (ct > 0.0)
        if np.any(st_sel):
            # apply selection:
            x = x[st_sel]
            ct = ct[st_sel]

            # get D:
            D = self.get_data(
                FV.D,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel]

            # get TI:
            ati = self.get_data(
                FV.AMB_TI,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )

            # get k:
            k = self.wake_k(
                FC.STATE_TARGET,
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
                amb_ti=ati,
            )[st_sel]

            ati = ati[st_sel]

            # calculate sigma:
            # beta = np.sqrt(0.5 * (1 + np.sqrt(1.0 - ct)) / np.sqrt(1.0 - ct))
            a = self.induction.ct2a(ct)
            beta = (1 - a) / (1 - 2 * a)
            epsilon = self.sbeta_factor * np.sqrt(beta)
            del a, beta

            alpha = self.c1 * ati
            beta = self.c2 * ati / np.sqrt(ct)

            # calculate sigma (eqn 4)
            sigma = D * (
                epsilon
                + k
                * ati
                / beta
                * (
                    np.sqrt((alpha + beta * x / D) ** 2 + 1)
                    - np.sqrt(1 + alpha**2)
                    - np.log(
                        (np.sqrt((alpha + beta * x / D) ** 2 + 1) + 1)
                        * alpha
                        / ((np.sqrt(1 + alpha**2) + 1) * (alpha + beta * x / D))
                    )
                )
            )

            del (
                x,
                alpha,
                beta,
                epsilon,
            )

            # calculate amplitude, same as in Bastankhah model (eqn 7)
            ct_eff = ct / (8 * (sigma / D) ** 2)
            ampld = np.maximum(-2 * self.induction.ct2a(ct_eff), -1)

        # case no targets:
        else:
            st_sel = np.zeros_like(x, dtype=bool)
            n_sp = np.sum(st_sel)
            ampld = np.zeros(n_sp, dtype=FC.DTYPE)
            sigma = np.zeros(n_sp, dtype=FC.DTYPE)

        return {FV.WS: (ampld, sigma)}, st_sel


class TurbOParkWakeIX(GaussianWakeModel):
    """
    The generalized TurbOPark wake model, integrating TI over the streamline.

    https://iopscience.iop.org/article/10.1088/1742-6596/2265/2/022063/pdf

    Attributes
    ----------
    dx: float
        The step size of the integral
    sbeta_factor: float
        Factor multiplying sbeta
    self_wake: bool
        Flag for considering only own wake in ti integral
    induction: foxes.core.AxialInductionModel or str
        The induction model
    ipars: dict
        Additional parameters for centreline integration
    wake_k: foxes.core.WakeK
        Handler for the wake growth parameter k

    :group: models.wake_models.wind

    """

    def __init__(
        self,
        superposition,
        dx,
        sbeta_factor=0.25,
        self_wake=True,
        induction="Madsen",
        ipars={},
        **wake_k,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind deficit superposition
        dx: float
            The step size of the integral
        sbeta_factor: float
            Factor multiplying sbeta
        self_wake: bool
            Flag for considering only own wake in ti integral
        induction: foxes.core.AxialInductionModel or str
            The induction model
        ipars: dict
            Additional parameters for centreline integration
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.WS: superposition})

        self.dx = dx
        self.sbeta_factor = sbeta_factor
        self.ipars = ipars
        self._tiwakes = None
        self.self_wake = self_wake
        self.induction = induction
        self.wake_k = WakeK(**wake_k)

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.WS]}, induction={iname}, dx={self.dx}, "
        s += self.wake_k.repr() + ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.wake_k, self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_rpoints, ...)

        """
        # find TI wake model:
        self._tiwakes = []
        for w in algo.wake_models.values():
            if w is not self:
                wdel = w.new_wake_deltas(algo, mdata, fdata, tdata)
                if self.wake_k.ti_var in wdel:
                    self._tiwakes.append(w)
        if self.wake_k.ti_var not in FV.amb2var and len(self._tiwakes) == 0:
            raise KeyError(
                f"Model '{self.name}': Missing wake model that computes wake delta for variable {self.wake_k.ti_var}"
            )

        return super().new_wake_deltas(algo, mdata, fdata, tdata)

    def calc_amplitude_sigma(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Calculate the amplitude and the sigma,
        both depend only on x (not on r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)

        Returns
        -------
        amsi: tuple
            The amplitude and sigma, both numpy.ndarray
            with shape (n_st_sel,)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        # get ct:
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # select targets:
        st_sel = (x > 1e-5) & (ct > 0.0)
        if np.any(st_sel):
            # apply selection:
            # x = x[st_sel]
            ct = ct[st_sel]

            # get D:
            D = self.get_data(
                FV.D,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel]

            # get k:
            k = self.wake_k(
                FC.STATE_TARGET,
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel]

            # calculate sigma:
            # beta = np.sqrt(0.5 * (1 + np.sqrt(1.0 - ct)) / np.sqrt(1.0 - ct))
            a = self.induction.ct2a(ct)
            beta = (1 - a) / (1 - 2 * a)
            epsilon = self.sbeta_factor * np.sqrt(beta)
            del a, beta

            # get TI by integration along centre line:
            ti_ix = algo.wake_frame.calc_centreline_integral(
                algo,
                mdata,
                fdata,
                downwind_index,
                [self.wake_k.ti_var],
                x,
                dx=self.dx,
                wake_models=self._tiwakes,
                self_wake=self.self_wake,
                **self.ipars,
            )[:, :, 0]

            # calculate sigma (eqn 1, plus epsilon from eqn 4 for x = 0)
            sigma = D * epsilon + k * ti_ix[st_sel]
            del x, epsilon

            # calculate amplitude, same as in Bastankhah model (eqn 7)
            ct_eff = ct / (8 * (sigma / D) ** 2)
            ampld = np.maximum(-2 * self.induction.ct2a(ct_eff), -1)

        # case no targets:
        else:
            st_sel = np.zeros_like(x, dtype=bool)
            n_sp = np.sum(st_sel)
            ampld = np.zeros(n_sp, dtype=FC.DTYPE)
            sigma = np.zeros(n_sp, dtype=FC.DTYPE)

        return {FV.WS: (ampld, sigma)}, st_sel

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().finalize(algo, verbosity)
        self._tiwakes = None



from foxes.core import TurbineType


class NullType(TurbineType):
    """
    A turbine type that does not compute any data.

    :group: models.turbine_types

    """

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return []

    def calculate(self, algo, mdata, fdata, st_sel):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: numpy.ndarray of bool
            The state-turbine selection,
            shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        return {}



import importlib
import importlib.util
import sys


def import_module(name, package=None, hint=None):
    """
    Imports a module dynamically.

    Parameters
    ----------
    name: str
        The module name
    package: str, optional
        The explicit package name, deduced from name
        if not given
    hint: str, optional
        Installation advice, in case the import fails

    Returns
    -------
    mdl: module
        The imnported package

    :group: utils

    """
    try:
        return importlib.import_module(name, package)
    except ModuleNotFoundError:
        mdl = name if package is None else f"{package}.{name}"
        hint = hint if hint is not None else f"pip install {name}"
        raise ModuleNotFoundError(f"Module '{mdl}' not found, maybe try '{hint}'")


def load_module(name, path):
    """
    Imports a module from file path

    Parameters
    ----------
    name: str
        The name of the module
    path: str
        The path to the python file

    Returns
    -------
    module:
        The module object

    :group: utils

    """
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)

    return module



"""
Algorithms define the main calculation routines.
"""

from .downwind.downwind import Downwind
from .iterative.iterative import Iterative
from .sequential.sequential import Sequential

from . import downwind
from . import iterative
from . import sequential



import numpy as np

from foxes.core import PartialWakesModel
import foxes.variables as FV
import foxes.constants as FC


class PartialSegregated(PartialWakesModel):
    """
    Add the averaged wake effects to the separately
    averaged ambient rotor results

    Attributes
    ----------
    rotor_model: foxes.core.RotorModel
        The rotor model, default is the one from the algorithm
    grotor: foxes.models.rotor_models.GridRotor
        The grid rotor model

    :group: models.partial_wakes

    """

    def __init__(self, rotor_model):
        """
        Constructor.

        Parameters
        ----------
        rotor_model: foxes.core.RotorModel
            The rotor model for wake averaging

        """
        super().__init__()

        self.rotor = rotor_model
        self.YZ = self.var("YZ")
        self.W = self.var(FV.WEIGHT)

    def __repr__(self):
        return f"{type(self).__name__}(rotor_model={self.rotor.name})"

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return super().sub_models() + [self.rotor]

    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        return (
            self.rotor.get_rotor_points(algo, mdata, fdata),
            self.rotor.rotor_point_weights(),
        )

    def finalize_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        amb_res,
        rpoint_weights,
        wake_deltas,
        wmodel,
        downwind_index,
    ):
        """
        Updates the wake_deltas at the selected target
        downwind index.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.Data
            The target point data
        amb_res: dict
            The ambient results at the target points
            of all rotors. Key: variable name, value
            np.ndarray of shape:
            (n_states, n_turbines, n_rotor_points)
        rpoint_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: np.ndarray of shape
            (n_states, n_turbines, n_tpoints)
        wmodel: foxes.core.WakeModel
            The wake model
        downwind_index: int
            The index in the downwind order

        Returns
        -------
        final_wake_deltas: dict
            The final wake deltas at the selected downwind
            turbines. Key: variable name, value: np.ndarray
            of shape (n_states, n_rotor_points)

        """
        n_states = fdata.n_states
        n_rotor_points = len(rpoint_weights)
        gweights = tdata[FC.TWEIGHTS]

        wdel = {v: d[:, downwind_index, None].copy() for v, d in wake_deltas.items()}

        if n_rotor_points == tdata.n_tpoints:
            ares = {v: d[:, downwind_index, None] for v, d in amb_res.items()}
        else:
            ares = {}
            for v, d in amb_res.items():
                ares[v] = np.zeros((n_states, 1, tdata.n_tpoints), dtype=FC.DTYPE)
                ares[v][:] = np.einsum("sp,p->s", d[:, downwind_index], rpoint_weights)[
                    :, None, None
                ]

        wmodel.finalize_wake_deltas(algo, mdata, fdata, ares, wdel)

        for v in wdel.keys():
            hdel = np.zeros((n_states, n_rotor_points), dtype=FC.DTYPE)
            hdel[:] = np.einsum("sp,p->s", wdel[v][:, 0], gweights)[:, None]
            wdel[v] = hdel

        return wdel



import numpy as np
import pandas as pd
from cycler import cycler
import matplotlib.pyplot as plt

from .output import Output
import foxes.variables as FV
import foxes.constants as FC


class FarmResultsEval(Output):
    """
    Evaluates farm results data.

    This sums over turbines and/or states,
    given the state-turbine farm_calc results.

    Attributes
    ----------
    results: xarray.Dataset
        The farm results

    :group: output

    """

    def __init__(self, farm_results):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results

        """
        self.results = farm_results

    def weinsum(self, rhs, *vars):
        """
        Calculates Einstein sum, adding weights
        as last argument to the given fields.

        It's all about treating NaN values.

        Parameters
        ----------
        rhs: str
            The right-hand side of the einsum expression.
            Convention: 's' for states, 't' for turbines
        vars: tuple of str or np.ndarray
            The variables mentioned in the expression,
            but without the obligatory weights that will
            be added at the end

        Returns
        -------
        result: np.ndarray
            The results array

        """
        nas = None
        fields = []
        for v in vars:
            if isinstance(v, str):
                fields.append(self.results[v].to_numpy())
            else:
                fields.append(v)
            if nas is None:
                nas = np.zeros_like(fields[-1], dtype=bool)
            nas = nas | np.isnan(fields[-1])

        inds = ["st" for v in fields] + ["st"]
        expr = ",".join(inds) + "->" + rhs

        if np.any(nas):
            sel = ~np.any(nas, axis=1)
            fields = [f[sel] for f in fields]

            weights0 = self.results[FV.WEIGHT].to_numpy()
            w0 = np.sum(weights0, axis=0)[None, :]
            weights = weights0[sel]
            w1 = np.sum(weights, axis=0)[None, :]
            weights *= w0 / w1
            fields.append(weights)

        else:
            fields.append(self.results[FV.WEIGHT].to_numpy())

        return np.einsum(expr, *fields)

    def reduce_states(self, vars_op):
        """
        Reduces the states dimension by some operation

        Parameters
        ----------
        vars_op: dict
            The operation per variable. Key: str, the variable
            name. Value: str, the operation, choices
            are: sum, mean, min, max.

        Returns
        -------
        data: pandas.DataFrame
            The results per turbine

        """
        n_turbines = self.results.sizes[FC.TURBINE]

        rdata = {}
        for v, op in vars_op.items():
            if op == "mean":
                rdata[v] = self.weinsum("t", v)
            elif op == "sum":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.sum(vdata, axis=0)
            elif op == "min":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.min(vdata, axis=0)
            elif op == "max":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.max(vdata, axis=0)
            elif op == "std":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.std(vdata, axis=0)
            else:
                raise KeyError(
                    f"Unknown operation '{op}' for variable '{v}'. Please choose: sum, mean, min, max"
                )

        data = pd.DataFrame(index=range(n_turbines), data=rdata)
        data.index.name = FC.TURBINE

        return data

    def reduce_turbines(self, vars_op):
        """
        Reduces the turbine dimension by some operation

        Parameters
        ----------
        vars_op: dict
            The operation per variable. Key: str, the variable
            name. Value: str, the operation, choices
            are: sum, mean, min, max.

        Returns
        -------
        data: pandas.DataFrame
            The results per state

        """
        states = self.results.coords[FC.STATE].to_numpy()

        rdata = {}
        for v, op in vars_op.items():
            if op == "mean":
                rdata[v] = self.weinsum("s", v)
            elif op == "sum":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.sum(vdata, axis=1)
            elif op == "min":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.min(vdata, axis=1)
            elif op == "max":
                vdata = self.results[v].to_numpy()
                rdata[v] = np.max(vdata, axis=1)
            else:
                raise KeyError(
                    f"Unknown operation '{op}' for variable '{v}'. Please choose: sum, mean, min, max"
                )

        data = pd.DataFrame(index=states, data=rdata)
        data.index.name = FC.STATE

        return data

    def reduce_all(self, states_op, turbines_op):
        """
        Reduces states and turbine dimension by some operation

        Parameters
        ----------
        states_op: dict
            The states contraction operations.
            Key: str, the variable name. Value:
            str, the operation, choices are:
            sum, mean, min, max.
        turbines_op: dict
            The turbines contraction operations.
            Key: str, the variable name. Value:
            str, the operation, choices are:
            sum, mean, min, max.

        Returns
        -------
        data: dict
            The fully contracted results

        """
        sdata = self.reduce_states(states_op)

        rdata = {}
        for v, op in turbines_op.items():
            vdata = sdata[v].to_numpy()
            if op == "mean":
                if states_op[v] == "mean":
                    rdata[v] = self.weinsum("", v)
                else:
                    vdata = sdata[v].to_numpy()
                    rdata[v] = self.weinsum("", vdata[None, :])
            elif op == "sum":
                vdata = sdata[v].to_numpy()
                rdata[v] = np.sum(vdata)
            elif op == "min":
                vdata = sdata[v].to_numpy()
                rdata[v] = np.min(vdata)
            elif op == "max":
                vdata = sdata[v].to_numpy()
                rdata[v] = np.max(vdata)
            else:
                raise KeyError(
                    f"Unknown operation '{op}' for variable '{v}'. Please choose: sum, mean, min, max"
                )

        return rdata

    def calc_states_mean(self, vars):
        """
        Calculates the mean wrt states.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per turbine

        """
        if isinstance(vars, str):
            return self.reduce_states({vars: "mean"})
        return self.reduce_states({v: "mean" for v in vars})

    def calc_states_sum(self, vars):
        """
        Calculates the sum wrt states.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per turbine

        """
        return self.reduce_states({v: "sum" for v in vars})

    def calc_states_std(self, vars):
        """
        Calculates the standard deviation wrt states.

        Args:
            vars (_type_): _description_

        Returns:
            _type_: _description_
        """

        return self.reduce_states({v: "std" for v in vars})

    def calc_turbine_mean(self, vars):
        """
        Calculates the mean wrt turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per state

        """
        return self.reduce_turbines({v: "mean" for v in vars})

    def calc_turbine_sum(self, vars):
        """
        Calculates the sum wrt turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per state

        """
        return self.reduce_turbines({v: "sum" for v in vars})

    def calc_farm_mean(self, vars):
        """
        Calculates the mean over states and turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: dict
            The fully contracted results

        """
        op = {v: "mean" for v in vars}
        return self.reduce_all(states_op=op, turbines_op=op)

    def calc_farm_sum(self, vars):
        """
        Calculates the sum over states and turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: dict
            The fully contracted results

        """
        op = {v: "sum" for v in vars}
        return self.reduce_all(states_op=op, turbines_op=op)

    def calc_mean_farm_power(self, ambient=False):
        """
        Calculates the mean total farm power.

        Parameters
        ----------
        ambient: bool
            Flag for ambient power

        Returns
        -------
        data: float
            The mean wind farm power

        """
        v = FV.P if not ambient else FV.AMB_P
        cdata = self.reduce_all(states_op={v: "mean"}, turbines_op={v: "sum"})
        return cdata[v]

    def calc_turbine_yield(
        self,
        algo=None,
        annual=False,
        ambient=False,
        hours=None,
        delta_t=None,
        P_unit_W=None,
    ):
        """
        Calculates the yield per turbine

        Parameters
        ----------
        algo: foxes.core.Algorithm, optional
            The algorithm, for P_nominal lookup
        annual: bool, optional
            Flag for returing annual results, by default False
        ambient: bool, optional
            Flag for ambient power, by default False
        hours: int, optional
            The duration time in hours, if not timeseries states
        delta_t: np.datetime64, optional
            The time delta step in case of time series data,
            by default automatically determined
        P_unit_W: float
            The power unit in Watts, 1000 for kW. Looked up
            in algorithm if not given

        Returns
        -------
        pandas.DataFrame
            A dataframe of yield values by turbine in GWh

        """
        if ambient:
            var_in = FV.AMB_P
            var_out = FV.AMB_YLD
        else:
            var_in = FV.P
            var_out = FV.YLD

        if algo is not None and P_unit_W is None:
            P_unit_W = np.array(
                [FC.P_UNITS[t.P_unit] for t in algo.farm_controller.turbine_types],
                dtype=FC.DTYPE,
            )[:, None]
        elif algo is None and P_unit_W is not None:
            pass
        else:
            raise KeyError("Expecting either 'algo' or 'P_unit_W'")

        # compute yield per turbine
        if np.issubdtype(self.results[FC.STATE].dtype, np.datetime64):
            if hours is not None:
                raise KeyError("Unexpected parameter 'hours' for timeseries data")
            times = self.results[FC.STATE].to_numpy()
            if delta_t is None:
                delta_t = times[-1] - times[-2]
            duration = times[-1] - times[0] + delta_t
            duration_seconds = np.int64(duration.astype(np.int64) / 1e9)
            duration_hours = duration_seconds / 3600
        elif hours is None and annual == True:
            duration_hours = 8760
        elif hours is None:
            raise ValueError(
                f"Expecting parameter 'hours' for non-timeseries data, or 'annual=True'"
            )
        else:
            duration_hours = hours
        yld = self.calc_states_mean(var_in) * duration_hours * P_unit_W / 1e9

        if annual:
            # convert to annual values
            yld *= 8760 / duration_hours

        yld.rename(columns={var_in: var_out}, inplace=True)
        return yld

    def add_capacity(self, algo=None, P_nom=None, ambient=False, verbosity=1):
        """
        Adds capacity to the farm results

        Parameters
        ----------
        algo: foxes.core.Algorithm, optional
            The algorithm, for nominal power calculation
        P_nom: list of float, optional
            Nominal power values for each turbine, if algo not given
        ambient: bool, optional
            Flag for calculating ambient capacity, by default False
        verbosity: int
            The verbosity level, 0 = silent

        """
        if ambient:
            var_in = FV.AMB_P
            var_out = FV.AMB_CAP
        else:
            var_in = FV.P
            var_out = FV.CAP

        # get results data for the vars variable (by state and turbine)
        vdata = self.results[var_in]

        if algo is not None and P_nom is None:
            P_nom = np.array(
                [t.P_nominal for t in algo.farm_controller.turbine_types],
                dtype=FC.DTYPE,
            )
        elif algo is None and P_nom is not None:
            P_nom = np.array(P_nom, dtype=FC.DTYPE)
        else:
            raise KeyError("Expecting either 'algo' or 'P_nom'")

        # add to farm results
        self.results[var_out] = vdata / P_nom[None, :]
        if verbosity > 0:
            if ambient:
                print("Ambient capacity added to farm results")
            else:
                print("Capacity added to farm results")

    def calc_farm_yield(self, turbine_yield=None, power_uncert=None, **kwargs):
        """
        Calculates yield, P75 and P90 at the farm level

        Parameters
        ----------
        turbine_yield: pandas.DataFrame, optional
            Yield values by turbine
        power_uncert: float, optional
            Uncertainty in the power value. Triggers
            P75 and P90 outputs
        kwargs: dict, optional
            Parameters for calc_turbine_yield(). Apply if
            turbine_yield is not given

        Returns
        -------
        farm_yield: float
            Farm yield result, same unit as turbine yield
        P75: float, optional
            The P75 value, same unit as turbine yield
        P90: float, optional
            The P90 value, same unit as turbine yield

        """
        if turbine_yield is None:
            yargs = dict(annual=True)
            yargs.update(kwargs)
            turbine_yield = self.calc_turbine_yield(**yargs)
        farm_yield = turbine_yield.sum()

        if power_uncert is not None:
            P75 = farm_yield * (1.0 - (0.675 * power_uncert))
            P90 = farm_yield * (1.0 - (1.282 * power_uncert))
            return farm_yield["YLD"], P75["YLD"], P90["YLD"]

        return farm_yield["YLD"]

    def add_efficiency(self, verbosity=1):
        """
        Adds efficiency to the farm results

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        P = self.results[FV.P]
        P0 = self.results[FV.AMB_P] + 1e-14
        self.results[FV.EFF] = P / P0  # add to farm results
        if verbosity:
            print("Efficiency added to farm results")

    def calc_farm_efficiency(self):
        """
        Calculates farm efficiency

        Returns
        -------
        eff: float
            The farm efficiency

        """
        P = self.calc_mean_farm_power()
        P0 = self.calc_mean_farm_power(ambient=True) + 1e-14
        return P / P0

    def gen_stdata(
        self,
        turbines,
        variable,
        fig=None,
        ax=None,
        figsize=None,
        legloc="lower right",
        animated=True,
        ret_im=True,
    ):
        """
        Generates state-turbine data,
        intended to be used in animations

        Parameters
        ----------
        turbines: list of int
            The turbines for which to scatter data
        variable: str
            The variable name
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        figsize: tuple, optional
            The figsize for plt.Figure
        legloc: str
            The legend location
        animated: bool
            Flag for animated output
        ret_im: bool
            Flag for image return,

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        im: List of matplotlib.collections.PathCollection, optional
            The scatter artists

        """

        if fig is None:
            hfig = plt.figure(figsize=figsize)
        else:
            hfig = fig
        if ax is None:
            hax = hfig.add_subplot(111)
        else:
            hax = ax

        hax.set_xlabel(f"State")
        hax.set_ylabel(variable)
        cc = cycler(color="bgrcmyk")

        data = self.results[variable].to_numpy()
        hasl = set()
        for si in range(len(data)):
            im = []
            hax.set_prop_cycle(cc)
            for ti in turbines:
                lbl = None if ti in hasl else f"Turbine {ti}"
                im += hax.plot(range(si), data[:si, ti], label=lbl, animated=animated)
                hasl.add(ti)

            hax.legend(loc=legloc)

            if ret_im:
                yield hfig, im
            else:
                yield hfig



from foxes.algorithms.downwind.downwind import Downwind

from foxes.core import FarmDataModelList
from foxes.utils import Dict
import foxes.variables as FV
from . import models as mdls


class Iterative(Downwind):
    """
    Iterative calculation of farm data.

    Attributes
    ----------
    max_it: int
        The maximal number of iterations
    conv_crit: foxes.algorithms.iterative.ConvCrit
        The convergence criteria
    prev_farm_results: xarray.Dataset
        Results from the previous iteration

    :group: algorithms.iterative

    """

    @classmethod
    def get_model(cls, name):
        """
        Get the algorithm specific model

        Parameters
        ----------
        name: str
            The model name

        Returns
        -------
        model: foxes.core.model
            The model

        """
        try:
            return getattr(mdls, name)
        except AttributeError:
            return super().get_model(name)

    def __init__(self, *args, max_it=None, conv_crit="default", mod_cutin={}, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Arguments for Downwind
        max_it: int, optional
            The maximal number of iterations
        conv_crit: foxes.algorithms.iterative.ConvCrit, optional
            The convergence criteria
        mod_cutin: dict, optional
            Parameters for cutin modification
        kwargs: dict, optional
            Keyword arguments for Downwind

        """
        super().__init__(*args, **kwargs)

        self.max_it = 2 + self.farm.n_turbines**2 if max_it is None else max_it
        self.conv_crit = (
            self.get_model("DefaultConv")() if conv_crit == "default" else conv_crit
        )
        self.prev_farm_results = None
        self._it = None
        self._mlist = None
        self._reamb = False
        self._urelax = None

        self._mod_cutin = dict(modify_ct=True, modify_P=False)
        self._mod_cutin.update(mod_cutin)

    def set_urelax(self, entry_point, **urel):
        """
        Sets under-relaxation parameters.

        Parameters
        ----------
        entry_point: str
            The entry point: first, pre_rotor, post_rotor,
            pre_wake, last
        urel: dict
            The variables and their under-relaxation values

        """
        if self.initialized:
            raise ValueError(f"Attempt to set_urelax after initialization")
        if self._urelax is None:
            self._urelax = Dict(
                first={},
                pre_rotor={},
                post_rotor={},
                pre_wake={},
                last={},
            )
        self._urelax[entry_point].update(urel)

    def initialize(self):
        """
        Initializes the algorithm.
        """
        super().initialize()
        if len(self._mod_cutin):
            for t in self.farm_controller.turbine_types:
                t.modify_cutin(**self._mod_cutin)

    @property
    def urelax(self):
        """
        Returns the under-relaxation parameters

        Returns
        -------
        urlx: foxes.utils.Dict
            The under-relaxation parameters

        """
        return self._urelax

    @property
    def iterations(self):
        """
        The current iteration number

        Returns
        -------
        it: int
            The current iteration number

        """
        return self._it

    def _collect_farm_models(
        self,
        outputs,
        calc_parameters,
        ambient,
    ):
        """
        Helper function that creates model list
        """
        if self._it == 0:
            self._mlist0, self._calc_pars0 = super()._collect_farm_models(
                outputs=False,
                calc_parameters=calc_parameters,
                ambient=ambient,
            )

            n = 0
            if self._urelax is not None:
                if len(self._urelax["first"]):
                    self._mlist0.insert(0, mdls.URelax(**self._urelax["first"]))
                    self._calc_pars0.insert(0, {})
                    self._reamb = True
                    n += 1

                if len(self._urelax["pre_rotor"]):
                    self._mlist0.insert(2 + n, mdls.URelax(**self._urelax["pre_rotor"]))
                    self._calc_pars0.insert(2 + n, {})
                    self._reamb = True
                    n += 1

                if len(self._urelax["post_rotor"]):
                    self._mlist0.insert(
                        4 + n, mdls.URelax(**self._urelax["post_rotor"])
                    )
                    self._calc_pars0.insert(4 + n, {})
                    self._reamb = True
                    n += 1

                if len(self._urelax["pre_wake"]):
                    self._mlist0.models[5 + n].urelax = mdls.URelax(
                        **self._urelax["pre_wake"]
                    )

                if len(self._urelax["last"]):
                    self._mlist0.append(mdls.URelax(**self._urelax["last"]))
                    self._calc_pars0.append({})

            return self._mlist0, self._calc_pars0

        elif ambient or self._reamb:
            return self._mlist0, self._calc_pars0

        else:
            # prepare:
            calc_pars = []
            mlist = FarmDataModelList(models=[])

            # do not rotate back from downwind order:
            if not self._final_run:

                # add under-relaxation during wake calculation:
                urelax = None
                if self._urelax is not None and len(self._urelax["pre_wake"]):
                    urelax = mdls.URelax(**self._urelax["pre_wake"])

                # add model that calculates wake effects:
                mlist.models.append(
                    self.get_model("FarmWakesCalculation")(urelax=urelax)
                )
                calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

                # add under-relaxation:
                if self._urelax is not None and len(self._urelax["last"]):
                    mlist.append(mdls.URelax(**self._urelax["last"]))
                    calc_pars.append({})

            # rotate back from downwind order:
            else:

                # add model that calculates wake effects:
                # mlist.models.append(self.get_model("FarmWakesCalculation")(urelax=None))
                # calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

                mlist.models.append(self.get_model("ReorderFarmOutput")(outputs))
                calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

            return mlist, calc_pars

    def _calc_farm_vars(self, mlist):
        """Helper function that gathers the farm variables"""
        if self._it == 0:
            super()._calc_farm_vars(mlist)

    def _run_farm_calc(self, mlist, *data, **kwargs):
        """Helper function for running the main farm calculation"""
        ir = (
            None
            if self.prev_farm_results is None
            else self.chunked(self.prev_farm_results)
        )
        return super()._run_farm_calc(mlist, *data, initial_results=ir, **kwargs)

    def calc_farm(self, finalize=True, **kwargs):
        """
        Calculate farm data.

        Parameters
        ----------
        finalize: bool
            Flag for finalization after calculation
        kwargs: dict, optional
            Arguments for calc_farm in the base class.

        Returns
        -------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)

        """
        outputs = kwargs.pop("outputs", self.DEFAULT_FARM_OUTPUTS)
        outputs = list(set(outputs + [FV.ORDER_SSEL, FV.ORDER_INV, FV.WEIGHT]))

        fres = None
        self._it = -1
        self._final_run = False
        while self._it < self.max_it:
            self._it += 1

            self.print(f"\nAlgorithm {self.name}: Iteration {self._it}\n", vlim=0)

            self.prev_farm_results = fres
            fres = super().calc_farm(outputs=None, finalize=False, **kwargs)

            if self.conv_crit is not None:
                conv = self.conv_crit.check_converged(
                    self, self.prev_farm_results, fres, verbosity=self.verbosity + 1
                )

                if conv:
                    self.print(
                        f"\nAlgorithm {self.name}: Convergence reached.\n", vlim=0
                    )
                    self.print("Starting final run")
                    self._final_run = True
                    fres = super().calc_farm(outputs=outputs, finalize=False, **kwargs)
                    break

            if self._it == 0:
                self.verbosity -= 1

        # finalize models:
        if finalize:
            self.print("\n")
            self.finalize()
            for m in self._mlist0.models:
                if m not in self.all_models():
                    m.finalize(self, self.verbosity)
            del self._mlist0, self._calc_pars0

        return fres



import numpy as np
import pandas as pd

import foxes.variables as FV
import foxes.constants as FC
from foxes.utils import wd2uv, uv2wd, TabWindroseAxes
from foxes.algorithms import Downwind
from foxes.core import WindFarm, Turbine
from foxes.models import ModelBook
from .output import Output


class RosePlotOutput(Output):
    """
     Class for rose plot creation

     Attributes
     ----------
     results: pandas.DataFrame
         The calculation results (farm or points)

    :group: output

    """

    def __init__(self, results):
        """
        Constructor.

        Parameters
        ----------
        results: xarray.Dataset
            The calculation results (farm or points)

        """
        dims = list(results.sizes.keys())
        if dims[1] == FC.TURBINE:
            self._rtype = FC.TURBINE
        elif dims[1] == FC.POINT:
            self._rtype = FC.POINT
        else:
            raise KeyError(
                f"Results dimension 1 is neither '{FC.TURBINE}' nor '{FC.POINT}': dims = {results.dims}"
            )

        self.results = results.to_dataframe()

    @classmethod
    def get_data_info(cls, dname):
        """
        Returns default description for a variable.

        Parameters
        ----------
        dname: str
            The variable name

        Returns
        -------
        title: str
            The long name of the variable
        legend: str
            The legend/axis text

        """

        if dname == FV.D:
            return "Rotor diameter", f"{FV.D} [m]"
        if dname == FV.H:
            return "Hub height", f"{FV.H} [m]"

        if dname == FV.WS:
            return "Wind speed", f"{FV.WS} [m/s]"
        if dname == FV.REWS:
            return "Rotor equivalent wind speed", f"{FV.REWS} [m/s]"
        if dname == FV.REWS2:
            return "Rotor equivalent wind speed (2nd moment)", f"{FV.REWS2} [m/s]"
        if dname == FV.REWS3:
            return "Rotor equivalent wind speed (3rd moment)", f"{FV.REWS3} [m/s]"
        if dname == FV.WD:
            return "Wind direction", f"{FV.WD} [deg]"
        if dname == FV.TI:
            return "Turbulence intensity", f"{FV.TI} [1]"
        if dname == FV.RHO:
            return "Air density", f"{FV.RHO} [kg/m3]"
        if dname == FV.CT:
            return "Thrust coefficient", f"{FV.CT} [1]"
        if dname == FV.P:
            return "Power", f"{FV.P} [kW]"
        if dname == FV.YAW:
            return "Yaw angle", f"{FV.YAW} [deg]"
        if dname == FV.YAWM:
            return "Yaw misalignment", f"{FV.YAWM} [deg]"

        if dname in FV.amb2var:
            title, legend = cls.get_data_info(FV.amb2var[dname])
            return f"Ambient {title.lower()}", f"AMB_{legend}"

        return dname, dname

    def get_data(
        self,
        sectors,
        var,
        var_bins,
        wd_var=FV.AMB_WD,
        turbine=None,
        point=None,
        start0=False,
    ):
        """
        Get pandas DataFrame with wind rose data.

        Parameters
        ----------
        sectors: int
            The number of wind direction sectors
        var: str
            The data variable name
        var_bins: list of float
            The variable bin seperation values
        wd_var: str, optional
            The wind direction variable name
        turbine: int, optional
            Only relevant in case of farm results.
            If None, mean over all turbines.
            Else, data from a single turbine
        point: int, optional
            Only relevant in case of point results.
            If None, mean over all points.
            Else, data from a single point
        start0: bool
            Flag for starting the first sector at
            zero degrees instead of minus half width

        Returns
        -------
        pd.DataFrame:
            The wind rose data

        """

        dwd = 360.0 / sectors
        wds = np.arange(0.0, 360.0, dwd)
        wdb = np.append(wds, 360) if start0 else np.arange(-dwd / 2, 360.0, dwd)
        lgd = f"interval_{var}"

        data = self.results[[wd_var, FV.WEIGHT]].copy()
        data[lgd] = self.results[var]
        uv = wd2uv(data[wd_var].to_numpy())
        data["u"] = uv[:, 0]
        data["v"] = uv[:, 1]

        data[FV.WEIGHT] *= 100
        data = data.rename(columns={FV.WEIGHT: "frequency"})

        el = turbine if self._rtype == FC.TURBINE else point
        if el is None:
            data = data.groupby(level=0).mean()
        else:
            sname = data.index.names[0]
            grp = data.reset_index().groupby(self._rtype)
            data = grp.get_group(el).set_index(sname)

        data["wd"] = uv2wd(data[["u", "v"]].to_numpy())
        data.drop(["u", "v"], axis=1, inplace=True)
        if not start0:
            data.loc[data["wd"] > 360.0 - dwd / 2, "wd"] -= 360.0

        data[wd_var] = pd.cut(data["wd"], wdb, labels=wds)
        data[lgd] = pd.cut(data[lgd], var_bins, right=False, include_lowest=True)

        grp = data[[wd_var, lgd, "frequency"]].groupby([wd_var, lgd], observed=False)
        data = grp.sum().reset_index()

        data[wd_var] = data[wd_var].astype(np.float64)
        data[lgd] = list(data[lgd])
        if start0:
            data[wd_var] += dwd / 2

        ii = pd.IntervalIndex(data[lgd])
        data[var] = ii.mid
        data[f"bin_min_{var}"] = ii.left
        data[f"bin_max_{var}"] = ii.right
        data[f"bin_min_{wd_var}"] = data[wd_var] - dwd / 2
        data[f"bin_max_{wd_var}"] = data[wd_var] + dwd / 2
        data["sector"] = (data[wd_var] / dwd).astype(int)

        data = data[
            [
                wd_var,
                var,
                "sector",
                f"bin_min_{wd_var}",
                f"bin_max_{wd_var}",
                f"bin_min_{var}",
                f"bin_max_{var}",
                lgd,
                "frequency",
            ]
        ]
        data.index.name = "bin"

        return data

    def get_figure(
        self,
        sectors,
        var,
        var_bins,
        wd_var=FV.AMB_WD,
        turbine=None,
        point=None,
        title=None,
        legend=None,
        design="bar",
        start0=False,
        fig=None,
        figsize=None,
        rect=None,
        ret_data=False,
        **kwargs,
    ):
        """
        Creates figure object

        Parameters
        ----------
        sectors: int
            The number of wind direction sectors
        var: str
            The data variable name
        var_bins: list of float
            The variable bin seperation values
        wd_var: str, optional
            The wind direction variable name
        turbine: int, optional
            Only relevant in case of farm results.
            If None, mean over all turbines.
            Else, data from a single turbine
        point: int, optional
            Only relevant in case of point results.
            If None, mean over all points.
            Else, data from a single point
        title. str, optional
            The title
        legend: str, optional
            The data legend string
        design: str
            The wind rose design: bar, contour, ...
        start0: bool
            Flag for starting the first sector at
            zero degrees instead of minus half width
        fig: matplotlib.Figure
            The figure to which to add an axis
        figsize: tuple, optional
            The figsize of the newly created figure
        rect: list, optional
            The rectangle of the figure which to fill,
            e.g. [0.1, 0.1, 0.8, 0.8]
        ret_data: bool
            Flag for returning wind rose data
        kwargs: dict, optional
            Additional arguments for TabWindroseAxes
            plot function

        Returns
        -------
        fig: matplotlib.pyplot.Figure
            The rose plot figure
        data: pd.DataFrame, optional
            The wind rose data

        """
        lg = legend
        if title is None or legend is None:
            ttl, lg = self.get_data_info(var)
        if title is not None:
            ttl = title

        wrdata = self.get_data(
            sectors=sectors,
            var=var,
            var_bins=var_bins,
            wd_var=wd_var,
            turbine=turbine,
            point=point,
            start0=start0,
        )

        ax = TabWindroseAxes.from_ax(fig=fig, rect=rect, figsize=figsize)
        fig = ax.get_figure()

        plfun = getattr(ax, design)
        plfun(
            direction=wrdata[wd_var].to_numpy(),
            var=wrdata[var].to_numpy(),
            weights=wrdata["frequency"].to_numpy(),
            bin_min_dir=np.sort(wrdata[f"bin_min_{wd_var}"].unique()),
            bin_min_var=np.sort(wrdata[f"bin_min_{var}"].unique()),
            bin_max_var=np.sort(wrdata[f"bin_max_{var}"].unique()),
            **kwargs,
        )
        ax.set_legend(title=lg)
        ax.set_title(ttl)

        if ret_data:
            return fig, wrdata
        else:
            return fig

    def write_figure(
        self,
        file_name,
        sectors,
        var,
        var_bins,
        ret_data=False,
        **kwargs,
    ):
        """
        Write rose plot to file

        Parameters
        ----------
        file_name: str
            Path to the output file
        sectors: int
            The number of wind direction sectors
        var: str
            The data variable name
        var_bins: list of float
            The variable bin seperation values
        ret_data: bool
            Flag for returning wind rose data
        kwargs: dict, optional
            Additional parameters for get_figure()

        Returns
        -------
        data: pd.DataFrame, optional
            The wind rose data

        """

        r = self.get_figure(
            sectors=sectors,
            var=var,
            var_bins=var_bins,
            ret_data=ret_data,
            **kwargs,
        )
        if ret_data:
            r[0].write_image(file_name)
            return r[1]
        else:
            r.write_image(file_name)


class StatesRosePlotOutput(RosePlotOutput):
    """
     Class for rose plot creation directly from states

     Parameters
     ----------
     states: foxes.core.States
         The states from which to compute the wind rose
     point: numpy.ndarray
         The evaluation point, shape: (3,)
     mbook: foxes.models.ModelBook, optional
         The model book

    :group: output

    """

    def __init__(self, states, point, mbook=None, ws_var=FV.AMB_REWS):
        farm = WindFarm()
        farm.add_turbine(
            Turbine(
                xy=point[:2],
                H=point[2],
                turbine_models=["null_type"],
            ),
            verbosity=0,
        )

        mbook = mbook if mbook is not None else ModelBook()
        algo = Downwind(farm, states, wake_models=[], mbook=mbook, verbosity=0)

        results = algo.calc_farm(ambient=True).rename_vars({ws_var: FV.AMB_WS})

        super().__init__(results)



from foxes.core import AxialInductionModel


class MadsenAxialInduction(AxialInductionModel):
    """
    Computes the induction factor through polynomial
    fit, extending validity for high ct values

    Notes
    -----
    Reference:
    Helge Aagaard Madsen, Torben Juul Larsen, Georg Raimund Pirrung, Ang Li, and Frederik Zahle
    "Implementation of the blade element momentum model on a polar grid and its aeroelastic load impact"
    https://doi.org/10.5194/wes-5-1-2020

    Attributes
    ----------
    k1: float
        Model coefficient
    k2: float
        Model coefficient
    k3: float
        Model coefficient

    :group: models.axial_induction

    """

    def __init__(self, k1=0.2460, k2=0.0586, k3=0.0883):
        """
        Constructor.

        Parameters
        ----------
        k1: float
            Model coefficient
        k2: float
            Model coefficient
        k3: float
            Model coefficient

        """
        super().__init__()
        self.k1 = k1
        self.k2 = k2
        self.k3 = k3

    def ct2a(self, ct):
        """
        Computes induction from ct

        Parameters
        ----------
        ct: numpy.ndarray or float
            The ct values

        Returns
        -------
        ct: numpy.ndarray or float
            The induction values

        """
        return self.k1 * ct + self.k2 * ct**2 + self.k3 * ct**3



from abc import abstractmethod
import numpy as np
from scipy.interpolate import interpn

from foxes.utils import all_subclasses
import foxes.constants as FC
import foxes.variables as FV

from .data import MData, FData, TData
from .model import Model


class WakeFrame(Model):
    """
    Abstract base class for wake frames.

    Wake frames translate global coordinates into
    wake frame coordinates, which are then evaluated
    by wake models.

    They are also responsible for the calculation of
    the turbine evaluation order.

    :group: core

    """

    @abstractmethod
    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        pass

    @abstractmethod
    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        pass

    def get_wake_modelling_data(
        self,
        algo,
        variable,
        downwind_index,
        fdata,
        tdata,
        target,
        states0=None,
        upcast=False,
    ):
        """
        Return data that is required for computing the
        wake from source turbines to evaluation points.

        Parameters
        ----------
        algo: foxes.core.Algorithm, optional
            The algorithm, needed for data from previous iteration
        variable: str
            The variable, serves as data key
        downwind_index: int, optional
            The index in the downwind order
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        target: str, optional
            The dimensions identifier for the output,
            FC.STATE_TURBINE, FC.STATE_TARGET,
            FC.STATE_TARGET_TPOINT
        states0: numpy.ndarray, optional
            The states of wake creation
        upcast: bool
            Flag for ensuring targets dimension,
            otherwise dimension 1 is entered

        Returns
        -------
        data: numpy.ndarray
            Data for wake modelling, shape:
            (n_states, n_turbines) or (n_states, n_target)
        dims: tuple
            The data dimensions

        """
        n_states = fdata.n_states
        s = np.s_[:] if states0 is None else states0

        if not upcast:
            if target == FC.STATE_TARGET_TPOINT:
                out = fdata[variable][s, downwind_index, None, None]
                dims = (FC.STATE, 1, 1)
            else:
                out = fdata[variable][s, downwind_index, None]
                dims = (FC.STATE, 1)
        elif target == FC.STATE_TURBINE:
            out = np.zeros((n_states, fdata.n_turbines), dtype=FC.DTYPE)
            out[:] = fdata[variable][s, downwind_index, None]
            dims = (FC.STATE, FC.TURBINE)
        elif target == FC.STATE_TARGET:
            out = np.zeros((n_states, tdata.n_targets), dtype=FC.DTYPE)
            out[:] = fdata[variable][s, downwind_index, None]
            dims = (FC.STATE, FC.TARGET)
        elif target == FC.STATE_TARGET_TPOINT:
            out = np.zeros((n_states, tdata.n_targets, tdata.n_tpoints), dtype=FC.DTYPE)
            out[:] = fdata[variable][s, downwind_index, None, None]
            dims = (FC.STATE, FC.TARGET, FC.TPOINT)
        else:
            raise ValueError(
                f"Unsupported target '{target}', expcting '{FC.STATE_TURBINE}', '{FC.STATE_TARGET}', {FC.STATE_TARGET_TPOINT}"
            )

        return out, dims

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        raise NotImplementedError(
            f"Wake frame '{self.name}': Centreline points requested but not implemented."
        )

    def calc_centreline_integral(
        self,
        algo,
        mdata,
        fdata,
        downwind_index,
        variables,
        x,
        dx,
        wake_models=None,
        self_wake=True,
        **ipars,
    ):
        """
        Integrates variables along the centreline.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        variables: list of str
            The variables to be integrated
        x: numpy.ndarray
            The wake frame x coordinates of the upper integral bounds,
            shape: (n_states, n_points)
        dx: float
            The step size of the integral
        wake_models: list of foxes.core.WakeModels
            The wake models to consider, default: from algo
        self_wake: bool
            Flag for considering only wake from states_source_turbine
        ipars: dict, optional
            Additional interpolation parameters

        Returns
        -------
        results: numpy.ndarray
            The integration results, shape: (n_states, n_points, n_vars)

        """
        # prepare:
        n_states, n_points = x.shape
        vrs = [FV.amb2var.get(v, v) for v in variables]
        n_vars = len(vrs)

        # calc evaluation points:
        xmin = 0.0
        xmax = np.nanmax(x)
        n_steps = int((xmax - xmin) / dx)
        if xmin + n_steps * dx < xmax:
            n_steps += 1
        n_ix = n_steps + 1
        xs = np.arange(xmin, xmin + n_ix * dx, dx)
        xpts = np.zeros((n_states, n_steps), dtype=FC.DTYPE)
        xpts[:] = xs[None, 1:]
        pts = self.get_centreline_points(algo, mdata, fdata, downwind_index, xpts)

        # run ambient calculation:
        tdata = TData.from_points(
            pts,
            data={
                v: np.full((n_states, n_steps, 1), np.nan, dtype=FC.DTYPE) for v in vrs
            },
            dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in vrs},
        )
        res = algo.states.calculate(algo, mdata, fdata, tdata)
        tdata.update(res)
        amb2var = algo.get_model("SetAmbPointResults")()
        amb2var.initialize(algo, verbosity=0)
        res = amb2var.calculate(algo, mdata, fdata, tdata)
        tdata.update(res)
        del res, amb2var

        # find out if all vars ambient:
        ambient = True
        for v in variables:
            if v not in FV.amb2var:
                ambient = False
                break

        # calc wakes:
        if not ambient:
            wcalc = algo.get_model("PointWakesCalculation")(wake_models=wake_models)
            wcalc.initialize(algo, verbosity=0)
            wsrc = downwind_index if self_wake else None
            res = wcalc.calculate(algo, mdata, fdata, tdata, downwind_index=wsrc)
            tdata.update(res)
            del wcalc, res

        # collect integration results:
        iresults = np.zeros((n_states, n_ix, n_vars), dtype=FC.DTYPE)
        for vi, v in enumerate(variables):
            for i in range(n_steps):
                iresults[:, i + 1, vi] = iresults[:, i, vi] + tdata[v][:, i, 0] * dx

        # interpolate to x of interest:
        qts = np.zeros((n_states, n_points, 2), dtype=FC.DTYPE)
        qts[:, :, 0] = np.arange(n_states)[:, None]
        qts[:, :, 1] = x
        qts = qts.reshape(n_states * n_points, 2)
        results = interpn(
            (np.arange(n_states), xs),
            iresults,
            qts,
            bounds_error=False,
            fill_value=0.0,
            **ipars,
        )

        return results.reshape(n_states, n_points, n_vars)

    @classmethod
    def new(cls, wframe_type, *args, **kwargs):
        """
        Run-time wake frame factory.

        Parameters
        ----------
        wframe_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if wframe_type is None:
            return None

        allc = all_subclasses(cls)
        found = wframe_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == wframe_type:
                    return scls(*args, **kwargs)

        else:
            estr = (
                "Wake frame type '{}' is not defined, available types are \n {}".format(
                    wframe_type, sorted([i.__name__ for i in allc])
                )
            )
            raise KeyError(estr)



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from foxes.core import VerticalProfile


class DataProfile(VerticalProfile):
    """
    A profile based on numerical data.

    Attributes
    ----------
    var: float
        The value
    data_z: numpy.ndarray
        The z values, shape: (n_z,)
    data_v: numpy.ndarray
        The variable values, shape: (n_z,)
    interp_pars: dict
        Additional parameters for interpolation

    :group: models.vertical_profiles

    """

    def __init__(
        self,
        data_source,
        variable,
        col_z=None,
        col_var=None,
        pd_read_pars={},
        **interp_pars
    ):
        """
        Constructor

        Parameters
        ----------
        data_source: str or numpy.ndarray or pandas.DataFrame
            The profile data
        variable: float
            The value
        col_z: str or int, optional
            The column of z data
        col_var: str or int, optional
            The column of variable data
        pd_read_pars: dict
            Additional parameters for pandas.read_csv()
        interp_pars: dict, optional
            Additional parameters for interpolation

        """
        super().__init__()
        self.var = variable
        self.interp_pars = interp_pars

        if isinstance(data_source, np.ndarray):
            col_z = col_z if col_z is not None else 0
            col_var = col_var if col_var is not None else -1
            self.data_z = data_source[col_z]
            self.data_v = data_source[col_var]
        else:
            if isinstance(data_source, pd.DataFrame):
                data = data_source
            else:
                data = pd.read_csv(data_source, **pd_read_pars)
            col_var = col_var if col_var is not None else variable
            self.data_v = data[col_var].to_numpy()
            if col_z is None:
                self.data_z = data.index.to_numpy()
            else:
                self.data_z = data[col_z].to_numpy()

        if not np.all(np.diff(self.data_z) > 0):
            inds = np.argsort(self.data_z)
            self.data_z = self.data_z[inds]
            self.data_v = self.data_v[inds]

    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        return []

    def calculate(self, data, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        data: dict
            The input data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """
        return interp1d(self.data_z, self.data_v, **self.interp_pars)(heights)



import numpy as np
from matplotlib.path import Path
from matplotlib.patches import PathPatch
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt

from .area_geometry import AreaGeometry


class ClosedPolygon(AreaGeometry):
    """
    This class represents a closed 2D polygon.

    Attributes
    ----------
    points: numpy.ndarray
        The polygon points
    poly: matplotlib.path.Path
        The closed polygon geometry

    :group: utils.geom2d

    """

    def __init__(self, points):
        """
        Constructor.

        Parameters
        ----------
        points: numpy.ndarray
            The polygon points, shape: (n_points, 2)

        """
        self.points = points

        if not np.all(points[0] == points[-1]):
            self.points = np.append(self.points, points[[0]], axis=0)

        self.poly = Path(self.points, closed=True)

        self._pathp = None

    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        return np.min(self.points, axis=0)

    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        return np.max(self.points, axis=0)

    def points_distance(self, points, return_nearest=False):
        """
        Calculates point distances wrt boundary.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)
        return_nearest: bool
            Flag for return of the nearest point on bundary

        Returns
        -------
        dist: numpy.ndarray
            The smallest distances to the boundary,
            shape: (n_points,)
        p_nearest: numpy.ndarray, optional
            The nearest points on the boundary, if
            return_nearest is True, shape: (n_points, 2)

        """

        dists = cdist(points, self.points[:-1])

        if return_nearest:
            mini = np.argmin(dists, axis=1)
            dists = np.take_along_axis(dists, mini[:, None], axis=1)[:, 0]
            minp = self.points[mini]
            del mini
        else:
            dists = np.min(dists, axis=1)

        for pi in range(len(self.points) - 1):
            pA = self.points[pi]
            pB = self.points[pi + 1]
            n = pB - pA
            d = np.linalg.norm(n)

            if d > 0:
                n /= d
                q = points - pA[None, :]
                x = np.einsum("pd,d->p", q, n)

                sel = (x > 0) & (x < d)
                if np.any(sel):
                    x = x[sel]
                    y2 = np.maximum(np.linalg.norm(q[sel], axis=1) ** 2 - x**2, 0.0)

                    dsel = dists[sel]
                    dists[sel] = np.minimum(dsel, np.sqrt(y2))

                    if return_nearest:
                        mini = np.argwhere(np.sqrt(y2) < dsel)
                        hminp = minp[sel]
                        hminp[mini] = pA[None, :] + x[mini, None] * n[None, :]
                        minp[sel] = hminp
                        del mini, hminp

                    del y2, dsel

                del x, sel

        if return_nearest:
            return dists, minp
        else:
            return dists

    def points_inside(self, points):
        """
        Tests if points are inside the geometry.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)

        Returns
        -------
        inside: numpy.ndarray
            True if point is inside, shape: (n_points,)

        """
        return self.poly.contains_points(points)

    def add_to_figure(
        self, ax, show_boundary=True, fill_mode=None, pars_boundary={}, pars_distance={}
    ):
        """
        Add image to (x,y) figure.

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The axis object
        show_boundary: bool
            Add the boundary line to the image
        fill_mode: str, optional
            Fill the area. Options:
            dist, dist_inside, dist_outside, inside_<color>,
            outside_<color>
        pars_boundary: dict
            Parameters for boundary plotting command
        pars_distance: dict
            Parameters for distance plotting command

        """
        if show_boundary:
            pars = dict(facecolor="none", edgecolor="darkblue", linewidth=1)
            pars.update(pars_boundary)

            pathpatch = PathPatch(self.poly, **pars)
            ax.add_patch(pathpatch)

        super().add_to_figure(
            ax, show_boundary, fill_mode, pars_boundary, pars_distance
        )


if __name__ == "__main__":
    points = np.array([[1.0, 1.0], [1.3, 6], [5.8, 6.2], [6.5, 0.8]])
    N = 500

    fig, ax = plt.subplots()
    g = ClosedPolygon(points)
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = ClosedPolygon(points)
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = ClosedPolygon(points).inverse()
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = ClosedPolygon(points).inverse()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)



import pandas as pd
import numpy as np

import foxes.variables as FV
import foxes.constants as FC


def create_random_abl_states(
    n_states, cols_minmax, var2col={}, mol_abs_range=(50.0, 5000.0), normalize=True
):
    """
    Create random abl states.

    Parameters
    ----------
    n_states: int
        The number of states
    cols_minmax: dict
        For each variable the min and max values,
        keys: variable name str, values: array_like
        with length 2
    var2col: dict, optional
        Mapping from variables to column names
    mol_abs_range: tuple
        Min and max of allowed MOL values, set to
        nan if exceeded (i.e., neutral stratification)
    normalize: bool
        Normalize weights to 1

    Returns
    -------
    data: pandas.DataFrame
        The created states data

    :group: input.states.create

    """

    data = pd.DataFrame(index=range(n_states))
    data.index.name = FC.STATE

    for v, mm in cols_minmax.items():
        data[v] = np.random.uniform(low=mm[0], high=mm[1], size=(n_states,)).astype(
            FC.DTYPE
        )

    cmol = var2col.get(FV.MOL, FV.MOL)
    if cmol in data and mol_abs_range is not None:
        sel = (data[cmol].abs() < mol_abs_range[0]) | (
            data[cmol].abs() > mol_abs_range[1]
        )
        data.loc[sel, cmol] = np.nan

    wcol = var2col.get(FV.WEIGHT, FV.WEIGHT)
    if wcol in data and normalize:
        data[wcol] /= data[wcol].sum()

    return data


def write_random_abl_states(
    file_path,
    n_states,
    cols_minmax,
    var2col={},
    mol_abs_range=(50.0, 5000.0),
    normalize=True,
    verbosity=1,
    digits="auto",
    **kwargs
):
    """
    Writes random abl states to file

    Parameters
    ----------
    file_path: str
        Path to the file
    n_states: int
        The number of states
    cols_minmax: dict
        For each variable the min and max values,
        keys: variable name str, values: array_like
        with length 2
    var2col: dict, optional
        Mapping from variables to column names
    mol_abs_range: tuple
        Min and max of allowed MOL values, set to
        nan if exceeded (i.e., neutral stratification)
    normalize: bool
        Normalize weights to 1
    verbosity: int
        The verbosity level, 0 = silent
    digits: int or auto
        The number of digits to be written
    kwargs: dict, optional
        Parameters for `pandas.DataFrame.to_csv`

    """

    if verbosity:
        print("Writing file", file_path)

    data = create_random_abl_states(
        n_states, cols_minmax, var2col, mol_abs_range, normalize
    )

    if digits is not None:
        hdigits = {c: 4 for c in cols_minmax.keys()}

        wcol = var2col.get(FV.WEIGHT, FV.WEIGHT)
        if wcol in cols_minmax:
            hdigits[wcol] = None

        mcol = var2col.get(FV.MOL, FV.MOL)
        if mcol in cols_minmax:
            hdigits[mcol] = 1

        tcol = var2col.get(FV.TI, FV.TI)
        if tcol in cols_minmax:
            hdigits[tcol] = 6

        if isinstance(digits, str) and digits == "auto":
            digits = hdigits
        else:
            digits = hdigits

    if digits is not None:
        for v, d in digits.items():
            if d is not None:
                data[v] = data[v].round(d)

    data.to_csv(file_path, **kwargs)



import numpy as np

from foxes.core import Turbine
import foxes.constants as FC


def add_grid(
    farm,
    xy_base,
    step_vectors,
    steps,
    indices=None,
    names=None,
    verbosity=1,
    **turbine_parameters
):
    """
    Add a regular grid of turbines.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    xy_base: numpy.ndarray
        The base point, shape: (2,)
    step_vectors: numpy.ndarray
        The two step vectors in x and y,
        respectively, shape: (2, 2)
    steps: array_like of int
        The steps in x, y. Length 2
    indices: list of int, optional
        The turbine indices
    names: list of str, optional
        The turbine names
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Parameters forwarded to `foxes.core.Turbine`

    :group: input.farm_layout

    """

    inds = list(np.ndindex(*steps))
    n_turbines = len(inds)

    xy_base = np.array(xy_base, dtype=FC.DTYPE)
    step_vectors = np.array(step_vectors, dtype=FC.DTYPE)

    for i in range(n_turbines):
        xi, yi = inds[i]
        farm.add_turbine(
            Turbine(
                xy=xy_base + xi * step_vectors[0] + yi * step_vectors[1],
                index=None if indices is None else indices[i],
                name=None if names is None else names[i],
                **turbine_parameters
            ),
            verbosity=verbosity,
        )



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colormaps
from matplotlib.colors import LinearSegmentedColormap

from foxes.input.states import SingleStateStates
from foxes.core import WindFarm
from foxes.algorithms import Downwind


class RotorPointPlot:
    """
    Visualizes rotor points and their weights.

    Attributes
    ----------
    rotor_model: foxes.core.RotorModel
        The rotor model
    algo: foxes.core.Algorithm, optional
        The algorithm

    :group: output

    """

    def __init__(self, rotor_model, algo=None):
        """
        Constructor.

        Parameters
        ----------
        rotor_model: foxes.core.RotorModel
            The rotor model
        algo: foxes.core.Algorithm, optional
            The algorithm

        """
        self.rotor_model = rotor_model
        self.algo = algo

        if self.algo is None:
            farm = WindFarm()
            states = SingleStateStates(ws=9, wd=270, ti=0.1, rho=1.225)
            self.algo = Downwind(farm, states, [])

    def get_point_figure(
        self,
        ax=None,
        fig=None,
        figsize=(5, 5),
        title=None,
        **kwargs,
    ):
        """
        Get a scatter plot of the rotor points.

        Parameters
        ----------
        ax: matplotlib.Axes, optional
            The plot axes
        fig: matplotlib.Figure, optional
            The figure object
        figsize: tuple
            The default figure size
        title: str, optional
            The plot title
        kwargs: dict, optional
            Additional arguments for pyplot.scatter

        Returns
        -------
        ax: matplotlib.Axes
            The plot axes

        """
        if fig is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
        else:
            ax = fig.axes[0] if ax is None else ax

        if not self.rotor_model.initialized:
            self.rotor_model.initialize(self.algo)

        points = self.rotor_model.design_points()
        weights = self.rotor_model.rotor_point_weights() * 100

        cmap = colormaps[kwargs.pop("cmap", "viridis_r")]
        wlist = np.sort(np.unique(weights))

        im = ax.scatter(points[:, 1], points[:, 2], c=weights, cmap=cmap, **kwargs)
        ax.add_patch(plt.Circle((0, 0), 1, color="black", fill=False, alpha=0.8))

        ax.set_xlabel("x/D")
        ax.set_ylabel("y/D")
        if title is None:
            title = str(self.rotor_model)
        if title != "":
            ax.set_title(title)
        ax.set_aspect("equal", adjustable="box")

        tlabels = [f"{wlist[-1]:.02f}"]
        for i in range(len(wlist) - 2, -1, -1):
            w = wlist[i]
            w1 = wlist[i + 1]
            if (w1 - w) / (wlist[-1] - wlist[0]) > 0.05:
                tlabels.insert(0, f"{w:.2f}")
            elif i == 0:
                tlabels[0] = ""
                tlabels.insert(0, f"{w:.2f}")
            else:
                tlabels.insert(0, "")

        cbar = fig.colorbar(im, ax=ax, label="Point weight [%]", shrink=0.8)
        cbar.set_ticks(ticks=wlist, labels=tlabels)

        return im



import matplotlib.pyplot as plt
import numpy as np

import foxes.constants as FC
from .output import Output


class StateTurbineMap(Output):
    """
    Creates heat maps with turbines on the one
    and states on the other axis.

    Attributes
    ----------
    results: xarray.Dataset
        The farm results

    :group: output

    """

    def __init__(self, farm_results):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results

        """
        self.results = farm_results

    def plot_map(
        self,
        variable,
        title=None,
        ax=None,
        figsize=None,
        **kwargs,
    ):
        """
        Plot the heat map for the selected variable.

        Parameters
        ----------
        variable: str
            The variable to plot
        title: str, optional
            The plot title
        ax: pyplot.Axis, optional
            The axis
        figsize: tuple
            The figsize argument for plt.subplots()
            in case ax is not provided
        kwargs: dict, optional
            Additional parameters for plt.pcolormesh()

        Returns
        -------
        ax: pyplot.Axis
            The plot axis

        """
        turbines = self.results[FC.TURBINE].to_numpy()
        states = self.results[FC.STATE].to_numpy()

        if ax is None:
            __, ax = plt.subplots(figsize=figsize)
        fig = ax.get_figure()

        ds = states[-1] - states[-2]
        states = np.append(states, states[-1] + ds)
        turbines = np.arange(len(turbines) + 1)

        y, x = np.meshgrid(states, turbines)
        z = self.results[variable].to_numpy()

        prgs = {"shading": "flat"}
        prgs.update(kwargs)

        c = ax.pcolormesh(x, y, z.T, **prgs)

        ax.set_xticks(turbines[:-1] + 0.5)
        ax.set_xticklabels(turbines[:-1])
        yt = ax.get_yticks()
        ytl = ax.get_yticklabels()
        ax.set_yticks(yt[:-1] + 0.5 * (yt[-1] - yt[-2]), ytl[:-1])
        if len(turbines) > 10:
            xt = ax.get_xticks()
            xtl = [None for t in xt]
            xtl[::5] = ax.get_xticklabels()[::5]
            ax.set_xticks(xt, xtl)
        fig.colorbar(c, ax=ax)

        t = title if title is not None else variable
        ax.set_title(t)
        ax.set_xlabel("Turbine index")
        ax.set_ylabel("State")

        return ax



import numpy as np


def sqrt_reg(x, x0=0.01):
    """
    A regularized sqrt function, producing
    non-zero values also for smallish negative x.

    Parameters
    ----------
    x: numpy.ndarray
        The x values to evaluate
    x0: float
        Parameter where to start the smoothing

    Returns
    -------
    out: numpy.ndarray
        The regularized sqrt(x) results

    :group: utils

    """
    b = x0 * (1 - np.log(x0))
    y = np.exp((x - b) / x0)
    return np.sqrt(np.where(x < x0, y, x))



import numpy as np

from .self_similar import SelfSimilar


class SelfSimilar2020(SelfSimilar):
    """
    The self-similar 2020 induction wake model
    from Troldborg and Meyer Forsting

    The individual wake effects are superposed linearly,
    without invoking a wake superposition model.

    Notes
    -----
    References:
    [1] Troldborg, Niels, and Alexander Raul Meyer Forsting.
    "A simple model of the wind turbine induction zone derived from numerical simulations."
    Wind Energy 20.12 (2017): 2011-2020.
    https://onlinelibrary.wiley.com/doi/full/10.1002/we.2137

    [2] Forsting, Alexander R. Meyer, et al.
    "On the accuracy of predicting wind-farm blockage."
    Renewable Energy (2023).
    https://www.sciencedirect.com/science/article/pii/S0960148123007620

    :group: models.wake_models.induction

    """

    def _a0(self, ct, x_R, gamma=1.1):
        """Helper function: define a0 with gamma factor, eqn 8 from [2]"""

        x_new = np.minimum(np.maximum(-1 * np.abs(x_R), -6), -1)
        c = (self._mu(x_new) - self._mu(-1)) / (self._mu(-6) - self._mu(-1))

        fg1 = -0.06489
        fg2 = 0.4911
        fg3 = 0.1577
        fg4 = 1.116
        far_gamma = fg1 * np.sin((ct - fg2) / fg3) + fg4

        ng1 = -1.381
        ng2 = 2.627
        ng3 = -1.524
        ng4 = 1.336
        near_gamma = ng1 * ct**3 + ng2 * ct**2 + (ng3 * ct) + ng4

        gamma = c * far_gamma + (1 - c) * near_gamma

        return self.induction.ct2a(gamma * ct)

    def _r_half(self, x_R):
        """Helper function: define induction zone half radius (eqn 13)"""
        return -0.672 * x_R + 0.4897



import dask
from abc import abstractmethod, ABCMeta
from copy import deepcopy
from dask.distributed import Client, LocalCluster
from dask.distributed import get_client
from dask.diagnostics import ProgressBar


class Runner(metaclass=ABCMeta):
    """
    Abstract base class for runners.

    :group: utils.runners

    """

    def __init__(self):
        self._initialized = False

    def initialize(self):
        """
        Initialize the runner
        """
        self._initialized = True

    @property
    def initialized(self):
        """
        Initialization flag

        Returns
        -------
        bool :
            Initialization flag

        """
        return self._initialized

    def __enter__(self):
        self.initialize()
        return self

    @abstractmethod
    def run(self, func, args=tuple(), kwargs={}):
        """
        Runs the given function.

        Parameters
        ----------
        func: Function
            The function to be run
        args: tuple
            The function arguments
        kwargs: dict
            The function keyword arguments

        Returns
        -------
        results: Any
            The functions return value

        """
        pass

    def finalize(self):
        """
        Finalize the runner
        """
        self._initialized = False

    def __exit__(self, exc_type, exc_value, traceback):
        self.finalize()


class DefaultRunner(Runner):
    """
    Class for default function execution.

    :group: utils.runners

    """

    def run(self, func, args=tuple(), kwargs={}):
        """
        Runs the given function.

        Parameters
        ----------
        func: Function
            The function to be run
        args: tuple
            The function arguments
        kwargs: dict
            The function keyword arguments

        Returns
        -------
        results: Any
            The functions return value

        """
        return func(*args, **kwargs)


class DaskRunner(Runner):
    """
    Class for function execution via dask

    Attributes
    ----------
    scheduler: str, optional
        The dask scheduler choice
    progress_bar: bool
        Flag for showing progress bar
    cluster_args: dict, optional
        Explicit arguments for the cluster setup
    client_args: dict, optional
        Explicit arguments for the client setup
    verbosity: int
        The verbosity level, 0 = silent

    :group: utils.runners

    """

    def __init__(
        self,
        scheduler=None,
        n_workers=None,
        threads_per_worker=None,
        processes=True,
        cluster_args=None,
        client_args={},
        progress_bar=True,
        verbosity=1,
    ):
        """
        Constructor.

        Parameters
        ----------
        scheduler: str, optional
            The dask scheduler choice
        n_workers: int, optional
            The number of workers for parallel run
        threads_per_worker: int, optional
            The number of threads per worker for parallel run
        progress_bar: bool
            Flag for showing progress bar
        cluster_args: dict, optional
            Explicit arguments for the cluster setup
        client_args: dict, optional
            Explicit arguments for the client setup
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().__init__()

        self.scheduler = scheduler
        self.client_args = client_args
        self.progress_bar = progress_bar
        self.verbosity = verbosity

        if cluster_args is None:
            self.cluster_args = dict(
                n_workers=n_workers,
                processes=processes,
                threads_per_worker=threads_per_worker,
            )
        elif n_workers is not None or threads_per_worker is not None:
            raise KeyError(
                "Cannot handle 'n_workers', 'threads_per_worker' arguments if 'cluster_args' are provided"
            )
        else:
            self.cluster_args = cluster_args

        if scheduler is None and (
            n_workers is not None
            or threads_per_worker is not None
            or cluster_args is not None
        ):
            self.scheduler = "distributed"

    @classmethod
    def is_distributed(cls):
        try:
            get_client()
            return True
        except ValueError:
            return False

    def initialize(self):
        """
        Initialize the runner
        """
        if self.scheduler is not None:
            dask.config.set(scheduler=self.scheduler)

        if self.scheduler == "dthreads":
            self.print("Launching local dask cluster..")

            cargs = deepcopy(self.cluster_args)
            del cargs["processes"]
            self._client = Client(processes=False, **self.client_args, **cargs)

            self.print(f"Dashboard: {self._client.dashboard_link}\n")

        elif self.scheduler == "distributed":
            self.print("Launching local dask cluster..")

            self._cluster = LocalCluster(**self.cluster_args)
            self._client = Client(self._cluster, **self.client_args)

            self.print(self._cluster)
            self.print(f"Dashboard: {self._client.dashboard_link}\n")

        elif self.scheduler == "slurm":
            from dask_jobqueue import SLURMCluster

            self.print("Launching dask cluster on HPC using SLURM..")

            cargs = deepcopy(self.cluster_args)
            nodes = cargs.pop("nodes", 1)

            self._cluster = SLURMCluster(**cargs)
            self._cluster.scale(jobs=nodes)
            self._client = Client(self._cluster, **self.client_args)

            self.print(self._cluster)
            self.print(f"Dashboard: {self._client.dashboard_link}\n")

        super().initialize()

    def print(self, *args, **kwargs):
        """
        Prints if verbosity is not zero
        """
        if self.verbosity > 0:
            print(*args, **kwargs)

    def run(self, func, args=tuple(), kwargs={}):
        """
        Runs the given function.

        Parameters
        ----------
        func: Function
            The function to be run
        args: tuple
            The function arguments
        kwargs: dict
            The function keyword arguments

        Returns
        -------
        results: Any
            The functions return value

        """
        if self.progress_bar:
            with ProgressBar():
                results = func(*args, **kwargs)
        else:
            results = func(*args, **kwargs)

        return results

    def finalize(self):
        """
        Finallize the runner
        """
        if self.scheduler in ["distributed", "slurm"]:
            self.print("\n\nShutting down dask cluster")
            self._client.close()
            self._cluster.close()

        dask.config.refresh()

        super().finalize()



def write_nc(ds, fpath, round="auto", complevel=5, verbosity=1, **kwargs):
    """
    Writes a dataset to netCFD file

    Parameters
    ----------
    fpath: str
        Path to the output file, should be nc
    round: dict or str, optional
        The rounding definitions, or auto for
        default settings
    complevel: int
        The compression level
    verbosity: int
        The verbosity level, 0 = silent
    kwargs: dict, optional
        Additional parameters for xarray.to_netcdf

    """

    if round is not None:
        for v in ds.coords.keys():
            if v in round:
                if verbosity > 1:
                    print(f"Rounding {v} to {round[v]} decimals")
                ds[v].data = ds[v].data.round(decimals=round[v])
        for v in ds.data_vars.keys():
            if v in round:
                if verbosity > 1:
                    print(f"Rounding {v} to {round[v]} decimals")
                ds[v].data = ds[v].data.round(decimals=round[v])

    if verbosity > 0:
        print("Writing file", fpath)

    enc = {k: {"zlib": True, "complevel": complevel} for k in ds.data_vars}
    ds.to_netcdf(fpath, encoding=enc, **kwargs)



from foxes.core import GroundModel
import foxes.variables as FV
import foxes.constants as FC


class WakeMirror(GroundModel):
    """
    Wake reflection from ground and/or other horizontal planes.

    Attributes
    ----------
    heights: list of float
        The reflection heights

    :group: models.ground_models

    """

    def __init__(self, heights):
        """
        Constructor.

        Parameters
        ----------
        heights: list of float
            The reflection heights

        """
        super().__init__()
        self.heights = heights

    def contribute_to_farm_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
        pwake,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model
        pwake: foxes.core.PartialWakesModel
            The partial wakes model

        """
        # prepare:
        hh = fdata[FV.H][:, downwind_index].copy()

        # contribution from main wake:
        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)
        wmodel.contribute(algo, mdata, fdata, tdata, downwind_index, wcoos, wake_deltas)

        # contribution from mirrors:
        tdata[FC.TARGETS] = tdata[FC.TARGETS].copy()  # making sure this is no ref
        for h in self.heights:

            fdata[FV.H][:, downwind_index] = hh + 2 * (h - hh)

            pwake.contribute(
                algo, mdata, fdata, tdata, downwind_index, wake_deltas, wmodel
            )

        # reset heights:
        fdata[FV.H][:, downwind_index] = hh

    def contribute_to_point_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model

        """
        # prepare:
        hh = fdata[FV.H][:, downwind_index].copy()

        # contribution from main wake:
        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)
        wmodel.contribute(algo, mdata, fdata, tdata, downwind_index, wcoos, wake_deltas)

        # contribution from mirrors:
        tdata[FC.TARGETS] = tdata[FC.TARGETS].copy()  # making sure this is no ref
        for h in self.heights:

            fdata[FV.H][:, downwind_index] = hh + 2 * (h - hh)

            wcoos = algo.wake_frame.get_wake_coos(
                algo, mdata, fdata, tdata, downwind_index
            )
            wmodel.contribute(
                algo, mdata, fdata, tdata, downwind_index, wcoos, wake_deltas
            )

        # reset heights:
        fdata[FV.H][:, downwind_index] = hh


class GroundMirror(WakeMirror):
    """
    Wake reflection from the ground.

    :group: models.ground_models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__(heights=[0.0])



from .farm_data_model import FarmDataModel


class FarmModel(FarmDataModel):
    """
    A FarmDataModel that operates on the farm level
    (in contrast to a TurbineModel, which allows
    application to a subset of turbines).

    :group: core

    """

    pass



import numpy as np

from foxes.core import FarmModel


class Turbine2FarmModel(FarmModel):
    """
    Wrapper that promotes turbine models
    into farm models, simply by selecting
    all turbines.

    Attributes
    ----------
    turbine_model: foxes.core.TurbineModel
        The turbine model

    :group: models.farm_models

    """

    def __init__(self, turbine_model):
        """
        Constructor.

        Parameters
        ----------
        turbine_model: foxes.core.TurbineModel
            The turbine model

        """
        super().__init__()
        self.turbine_model = turbine_model
        self.name = turbine_model.name + "_t2f"

    def __repr__(self):
        return f"{type(self).__name__}({self.turbine_model})"

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.turbine_model]

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.turbine_model.output_farm_vars(algo)

    def calculate(self, algo, mdata, fdata, **parameters):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        **parameters: dict, optional
            Init parameters forwarded to the turbine model

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        s = np.ones((algo.n_states, algo.n_turbines), dtype=bool)
        return self.turbine_model.calculate(algo, mdata, fdata, st_sel=s, **parameters)
